
import sys, os
from pathlib import Path

import numpy as np

import torch
from torch import nn
from torch.nn import functional as F

import omnifig as fig

from .. import util


@fig.AutoComponent('flip')
class RandomFlip(nn.Module):

	def __init__(self, horizontal=None, vertical=None):
		super().__init__()

		self.horizontal = horizontal
		assert horizontal is None or 0 <= horizontal <= 1
		self.vertical = vertical
		assert vertical is None or 0 <= vertical <= 1

	def forward(self, imgs):

		if self.horizontal is not None and self.horizontal > 0 and np.random.rand() < self.horizontal:
			# print('h')
			imgs = imgs.flip(-1)

		if self.vertical is not None and self.vertical > 0 and np.random.rand() < self.vertical:
			# print('v')
			imgs = imgs.flip(-2)

		return imgs


@fig.AutoComponent('random-orientation')
class RandomOrientation(nn.Module):

	def forward(self, imgs, x=None):

		if x is None:
			x = np.random.randint(8)

		if x < 4:
			# print('h')
			imgs = imgs.flip(-1) # horizontal pos
			if x < 2:
				# print('t')
				imgs = imgs.transpose(-1, -2)
				if x < 1:
					# print('h')
					imgs = imgs.flip(-1)  # horizontal pos
			elif x < 3:
				# print('v')
				imgs = imgs.flip(-2)  # vertical pos

		elif x < 6:
			# print('v')
			imgs = imgs.flip(-2)  # vertical pos
			if x < 5:
				# print('t')
				imgs = imgs.transpose(-1, -2)

		elif x < 7:
			# print('t')
			imgs = imgs.transpose(-1, -2)

		# else:
		# print('i')

		return imgs

@fig.AutoComponent('image-transform')
class Image_Transform(nn.Module):
	def __init__(self, prob=None,
	             flip_h=False, flip_w=False, rot_90=False,
	             rot=False, scale=None, offset=None, flip=None,
	             brightness=None, contrast=None, noise=None,
	             src_bg=True):

		super().__init__()

		if not flip_h and not flip_w and rot_90:
			raise NotImplementedError

		num = 0

		num += int(flip_h)

		num += int(flip_w)

		self.flip_h = flip_h
		self.flip_w = flip_w

		num += int(rot)
		contin = rot

		if scale is not None and scale > 0:
			num += 1
			contin = True

		if offset is not None and offset > 0:
			num += 1
			contin = True

		if contin and (flip_h or flip_w) and flip is None:
			flip = 0.5

		if brightness is not None and brightness > 0:
			num += 1

		if contrast is not None and contrast > 0:
			num += 1

		if noise is not None and noise > 0:
			num += 1

		if prob is not None:
			assert 0 < prob < 1
		self.prob = prob

		if num == 0:
			print('WARNING: no transforms selected')

		self.num = num
		self.contin = contin
		self.src_bg = src_bg
		self.orient = RandomOrientation() if flip_w and flip_h and rot_90 else None
		if self.orient is None and rot_90:
			raise NotImplementedError

		if flip is None and (flip_h or flip_w):
			flip = 0.5

		self.rot = rot
		assert scale is None or 0<scale<2, f'val: {scale}'
		self.scale = scale
		assert offset is None or 0<offset<2, f'val: {offset}'
		self.offset = offset

		assert flip is None or 0<=flip<=1, f'val: {flip}'
		self.flip = flip

		assert brightness is None or 0<=brightness<=1, f'val: {brightness}'
		self.brightness = brightness
		assert contrast is None or 0<=contrast<=1, f'val: {contrast}'
		self.contrast = contrast
		assert noise is None or 0<=noise<=1, f'val: {noise}'
		self.noise = noise

	def __len__(self):
		return self.num

	def gen_vals(self, mag, N, device='cpu'):
		if mag is None or mag <= 0:
			return torch.ones(N, device=device)
		return torch.rand(N, device=device)*2*mag + 1 - mag

	def forward(self, x):

		if len(self) and (self.prob is None or np.random.rand() < self.prob):

			B, C, H, W = x.size()

			if self.contin:

				if self.src_bg:
					alpha = torch.ones(B, 1, H, W, device=x.device)
					x = torch.cat([x, alpha], 1)
					initial = x

				thetas = torch.rand(B) * 2 * np.pi if self.rot else torch.zeros(B)
				xy = self.gen_vals(self.offset, B*2).sub(1).view(B, 2, 1)
				scales = self.gen_vals(self.scale, B).view(B,1,1)

				se2s = util.se2_tfm(util.rots_2d(thetas)*scales, xy).to(x.device)

				x = F.grid_sample(x, F.affine_grid(se2s, x.size()))

				if self.src_bg:
					mask = (x[:, -1:] > 0.99).float()
					x = x * mask + initial * (1 - mask)
					x = x[:, :-1]

				if self.flip is not None and self.flip > 0:
					sel = (torch.rand(B, 1, 1, 1) < self.flip).float().to(x.device)
					x = sel * x.flip(-1) + (1 - sel) * x

			elif self.orient is not None:
				x = self.orient(x)

			elif self.flip_h or self.flip_w:
				if self.flip_h and self.flip_w:
					v = np.random.rand()
					if v < .6666:
						x = x.flip(-1)
					if v > .3333:
						x = x.flip(-2)
				else:
					x = x.flip(-1 if self.flip_h else -2)

			if self.contrast is not None and self.contrast > 0:
				contrast = self.gen_vals(self.contrast, B, device=x.device).view(B,1,1,1)

				x = contrast * x + (1 - contrast) * x.mean(1, keepdim=True)

			if self.brightness is not None and self.brightness > 0:
				brightness = self.gen_vals(self.brightness, B, device=x.device).view(B,1,1,1)

				x = brightness * x

			if self.noise is not None and self.noise > 0:
				x = x + torch.randn_like(x)*self.noise

			x = x.clamp(min=0, max=1)

		return x




from . import util
from . import op
from . import data
from . import eval
# from . import sim

from . import models
# from .framework import * # TODO: cleanup
from .op.framework import FunctionBase, Model, Encodable, Decodable, Generative, Recordable, \
	Evaluatable, Visualizable, Function, Optimizable

from omnibelt import get_printer

prt = get_printer(__name__)

try:
	from . import legacy
except:
	prt.info('Failed to important legacy models')

import os
__info__ = {'__file__':os.path.join(os.path.abspath(os.path.dirname(__file__)), '_info.py')}
with open(os.path.join(os.path.abspath(os.path.dirname(__file__)), '_info.py'), 'r') as f:
	exec(f.read(), __info__)
del os
del __info__['__file__']
__author__ = __info__['author']
__version__ = __info__['version']




import sys, os, time
import traceback
from tqdm import tqdm
from tqdm.notebook import tqdm as tqdm_notebook
import math
import torch
import numpy as np
#from itertools import imap
from collections import deque, Mapping

from omnibelt import Simple_Child, Proper_Child, get_now, create_dir, save_yaml, load_yaml, Registry, Singleton

import omnifig as fig

from humpack import adict, tlist, tset, Table, TreeSpace

import torch.nn as nn
import torch.multiprocessing as mp
# from torch.utils.data.dataloader import ExceptionWrapper
import random

from .farming import make_ghost

FD_PATH = os.path.dirname(os.path.dirname(__file__))

# fig.register_config_dir(os.path.join(os.path.dirname(FD_PATH), 'config'))
fig.register_config('origin', os.path.join(os.path.dirname(FD_PATH), 'config', 'origin.yaml'))

DEFAULT_DATA_PATH = os.path.join(os.path.dirname(FD_PATH),'local_data')
DEFAULT_SAVE_PATH = os.path.join(os.path.dirname(FD_PATH),'trained_nets')


def to_np(tensor):
	return tensor.detach().cpu().numpy()

def split_vals(a, groups):
	idx = np.random.choice(np.arange(len(groups)), size=len(a), replace=True, p=groups)
	return [a[g==idx] for g in range(len(groups))]

def count_parameters(model):
	return sum(p.numel() for p in model.parameters())

def create_param(*sizes, requires_grad=True):
	t = torch.empty(*sizes)
	nn.init.xavier_normal_(t)
	return nn.Parameter(t, requires_grad=requires_grad)

@fig.Component('progress-bar')
class Progress_Bar(Singleton):
	def __init__(self, A):
		ptype = A.pull('display-on', 'cmd')
		
		self._pbar_cls = tqdm_notebook if ptype in {'notebook', 'jupyter'} else tqdm
		self.pbar = None
		self.pbariter = None
		
		self.default_limit = A.pull('limit', None)
		
		self.pause_state = None
		self._val = None
		self._desc = None
		self._limit = None
	
	def pause(self):
		self.pause_state = {'limit':self._limit, 'initial':self._val, 'desc':self._desc}
		self.reset()
	
	def unpause(self):
		if self.pause_state is not None:
			self.reset()
			self.init_pbar(**self.pause_state)
			self.pause_state = None
	
	def init_pbar(self, itr=None, limit=None, initial=0, desc=None, total=None, **kwargs):
		self.reset()
		if self._pbar_cls is not None:
			if itr is not None:
				try:
					limit = len(itr)
				except TypeError:
					pass
			if limit is None:
				limit = self.default_limit
			if total is None:
				total = limit
			self._limit = total
			self._val = initial
			self._desc = desc
			self.pbar = self._pbar_cls(itr, total=total, initial=initial, desc=desc, **kwargs)
			# if itr is not None:
			# 	self.pbariter = iter(self.pbar)
	
	def update(self, n=1, desc=None):
		if self.pbar is not None:
			self._desc = desc
			self._val += n
			if desc is not None:
				self.set_description(desc)
			if self.pbariter is not None:
				return next(self.pbariter)
			else:
				self.pbar.update(n=n)
	
	def __call__(self, itr=None, **kwargs):
		self.init_pbar(itr=itr, **kwargs)
		return self
	
	def __iter__(self):
		if self.pbariter is None and self.pbar is not None:
			self.pbariter = iter(self.pbar)
		return self
	
	def __next__(self):
		return self.update()
	
	def close(self):
		self.reset()
	
	def reset(self):
		if self.pbar is not None:
			try:
				self.pbar.close()
			except TypeError:
				pass
			print('\r', end='')
		self.pbar = None
		self.pbariter = None
	
	def set_description(self, desc):
		if self.pbar is not None:
			self.pbar.set_description(desc)





import sys, os
from pathlib import Path
import time
import socket
import random

from collections import OrderedDict

import humpack as hp

from omnibelt import load_yaml, save_yaml, get_now, create_dir, get_printer, unspecified_argument

import omnifig as fig
from omnifig import Configurable
from omnifig.errors import MissingParameterError, MissingComponentError


prt = get_printer(__name__)


_warn_savedir = True
def get_save_dir(A=None, silent=False):
	if A is not None:
		root = A.pull('saveroot', None, silent=silent)
		if root is not None:
			return Path(root)
	
	if 'OMNILEARN_SAVE_DIR' in os.environ:
		return Path(os.environ['OMNILEARN_SAVE_DIR'])
	
	root = Path(os.getcwd())
	root = root / 'trained_nets'
	root.mkdir(exist_ok=True)
	
	global _warn_savedir
	if _warn_savedir:
		prt.warning(f'No savedir found (specify with "OMNILEARN_SAVE_DIR" env variable), '
		            f'now using {str(root)}')
		_warn_savedir = False
	
	return root



class RunNotFoundError(Exception):
	def __init__(self, name):
		super().__init__(f'Run not found: {name}')



# class Validation_Flag(Exception):
# 	pass
#
# class Checkpoint_Flag(Exception):
# 	pass
#
# class NoOverwriteError(Exception):
# 	pass


@fig.Script('load-run', description='Load a new or existing run') # the script just loads a run
def load_run(A):
	
	silent = A.pull('silent', False, silent=True)
	
	override = A.pull('override', None, raw=True, silent=True)
	
	name = A.pull('path', '<>load', '<>resume', None, silent=silent)
	if name is not None:
		path = Path(name)
		
		if not path.is_dir():
			saveroot = get_save_dir(A)
			path = saveroot / path
		
		if not path.is_dir():
			raise RunNotFoundError(name)
		
		A = fig.get_config(str(path))
		if override is not None:
			A.update(override)
		A.push('path', str(path))
	
	if 'run' in A:
		A = A.sub('run')
	
	run_type = A.push('_type', 'run', overwrite=False, silent=True)
	if not fig.has_component(run_type):
		prt.error(f'Run type "{run_type}" not found, using default instead')
		A.push('_type', 'run')
		
	return A.pull_self()
	
	
@fig.Component('run')
class Run(Configurable):
	'''
	Holds all the data and functions to load, save, train, and evaluate runs.
	Runs include the model, datasets, dataloaders, the logger, and stats
	'''
	def __init__(self, A, name=unspecified_argument, path=unspecified_argument,
	             silent=None, invisible=None, use_root=None, **kwargs):
		
		if silent is None:
			silent = A.pull('silent', silent)

		if invisible is None:
			invisible = A.pull('invisible', False, silent=silent)

		if use_root is None:
			use_root = A.pull('use_config_root', True, silent=silent)
			
		if path is unspecified_argument:
			path = A.pull('path', '<>load', '<>resume', None, silent=silent)
		
		if name is unspecified_argument:
			name = A.pull('name', None, silent=silent) if path is None else None
		
		super().__init__(A, **kwargs)
		self.silent = silent
		self.invisible = invisible
		if use_root:
			A = A.get_root()
		self.A = A
		
		self.name = name
		
		if path is not None:
			A.push('path', path, silent=silent)
			path = Path(path)
		self.path = path
		self.novel = path is None
		
		if not invisible:
			self._setup_storage(A)

		self._prep(A)
		
		self.purge() # reset payload objs

		self.manual_prep()

	def manual_prep(self):
		pass
  
	def __repr__(self):
		return f'{self.__class__.__name__.upper()}:{self.get_name()}'
	
	def __str__(self):
		return self.get_name()
	
	def _prep(self, A):
		if 'clock' not in A:
			A.push('clock._type', 'clock', overwrite=False, silent=True)
		self.clock = A.pull('clock', ref=True)

		A.push('records._type', 'records', overwrite=False, silent=True)
		self.records = A.pull('records', ref=True)

	
	def purge(self): # remove any potentially large objects to free memory
		self.dataset = None
		self.model = None
		
		self.loader = None
		self._loader = None
		
		self.batch = None
		self.out = None
		
		self.clock.clear()
		self.records.purge()
		
		self.checkpointer = None
		self.validation = None
		self.vizualizer = None
		
		self.results = {}
	
		self._started = False
		
	def _setup_storage(self, A):
		if self.path is None:
			self._create_storage(A)
		else:
			self._load_storage(A)
		
	def _create_storage(self, A):
		
		name = self.get_name()
		
		if name is None:
			name = self._gen_name(A)
		
		logdate = A.pull('name-include-date', True)
		if logdate:
			now = get_now()
			name = f'{name}_{now}'
		
		self.name = name
		
		path = self.get_path()
		
		if path is None and not self.invisible:
			saveroot = get_save_dir(A, silent=self.silent)

			path = saveroot / self.name
			
			unique = A.pull('unique-name', True)
			if unique and self.novel and path.is_dir():
				idx = 2
				nameroot = self.name
				while path.is_dir():
					name = f'{nameroot}_{idx}'
					path = saveroot / name
				self.name = name
			
			path.mkdir(exist_ok=True)
			if self.novel:
				A.push('path', str(path))
		
		self.path = path
	
	def _load_storage(self, A):
		self.name = self.path.stem
		
		num = A.pull('ckpt-num', None, silent=self.silent)
		best = A.pull('best', False, silent=self.silent)
		if best:
			raise NotImplementedError
		last = A.pull('last', not best and num is None, silent=self.silent)
		
		ckpts = {int(path.stem[4:]):path for path in self.path.glob('ckpt*') if path.is_dir()}
		nums = sorted(ckpts.keys())
		
		if not len(nums):
			return
		
		ckpt = ckpts[nums[-1]] if last or num is None or num not in ckpts else ckpts[num]
		
		ckpt = str(ckpt)
		A.push('dataset._load-ckpt', ckpt, overwrite=True, silent=self.silent)
		A.push('model._load-ckpt', ckpt, overwrite=True, silent=self.silent)
		A.push('records._load-ckpt', ckpt, overwrite=True, silent=self.silent)
		A.push('clock._load-ckpt', ckpt, overwrite=True, silent=self.silent)
	
	# endregion

	def _gen_name(self, A):
		return 'test'
	
	@staticmethod
	def wrap_pull(A, keep_instance=True, **kwargs):
		A.begin()
		
		obj = A.pull_self()
		
		store_keys = A.pull('store_keys', [], silent=True)
		
		stored = {}
		
		if len(store_keys):
			
			for key in store_keys:
				try:
					val = A.pull(key, silent=True)
				except MissingParameterError:
					pass
				else:
					stored[key] = val
		
		A.abort()
		
		for key, val in stored.items():
			A.push(key, val, silent=True, force_root=True)
			
		if A.pull('keep_instance', keep_instance, silent=True):
			A.push('__obj', obj, silent=True)
		
		return obj
	
	def create_dataset(self, A=None, **meta):
		if A is None:
			A = self.get_config().sub('dataset')
		return self.wrap_pull(A, **meta)
	def create_model(self, A=None, **meta):
		if A is None:
			A = self.get_config().sub('model')
		return self.wrap_pull(A, **meta)
	def create_records(self, A=None, **meta):
		if A is None:
			A = self.get_config().sub('records')
		return A.pull_self()
	
	def create_loader(self, mode='train', **kwargs):
		dataset = self.get_dataset()
		dataset.switch_to(mode)
		return dataset.to_loader(**kwargs)
		
	def create_results(self, A=None, **meta):
		raise NotImplementedError # TODO
	
	def get_name(self):
		return self.name
	
	def get_description(self, ticks=None):
		if ticks is None:
			ticks = self.clock.get_time()
		
		tm = time.strftime("%H:%M:%S")
		
		progress = self.model.get_description()
		limit = self.clock.get_limit()
		
		mode = 'train'
		epochs = self.records['total_epochs'][mode]
		
		return f'[ {tm} ] {mode}:{epochs} {ticks}/{limit} {progress}'
	
	# region "Payload" objects - not loaded automatically

	def get_config(self):
		return self.A
	
	def get_clock(self):
		return self.clock

	def get_dataset(self, **meta):
		if self.dataset is None:
			self.dataset = self.create_dataset(**meta)
		return self.dataset
	def get_model(self, **meta):
		if self.model is None:
			self.model = self.create_model(**meta)
		return self.model
	def get_records(self, **meta):
		if self.records is None:
			self.records = self.create_records(**meta)
		return self.records
	
	def get_loader(self, mode=None, **kwargs):
		return self.get_dataset().get_loader(mode, **kwargs)
	
	def get_results(self, ident, path=None, remove_ext=True, **kwargs): # you must specify which results (ident used when results were created)
		
		if ident in self.results:
			return self.results[ident]

		fixed = ident.split('.')[0] if remove_ext else ident

		if fixed in self.results:
			return self.results[fixed]

		self.results[fixed] = self._load_results(name=ident, path=path, **kwargs)

		return self.results[fixed]


	def update_results(self, ident, data, path=None, overwrite=False, remove_ext=True):

		fixed = ident.split('.')[0] if remove_ext else ident

		if self.has_results(fixed, path=path) and not overwrite:
			old = self.get_results(fixed, path=path)
			old.update(data)
			data = old

		self._save_results(data, name=fixed, path=path)
		
		if fixed in self.results:
			self.results[fixed] = data
			
		return data


	def has_results(self, ident, path=None, ext=None):
		fixed = ident.split('.')[0] if ext is not None else ident
		
		if fixed in self.results:
			return True
		
		if ext is not None:
			fixed = f'{fixed}.{ext}'
		
		if path is None:
			path = self.get_path()
		return (path/fixed).is_file()

	# endregion
	
	# region Signals

	def get_path(self):
		return self.path

	def get_batch(self):
		if self.batch is None:
			self.batch = self.get_dataset().get_batch()
		return self.batch
	
	def get_output(self):
		if self.out is None:
			self.out = self.get_model().step(self.get_batch())
		return self.out

	# endregion
	
	# region Training Phases
	
	def take_steps(self, n=1, complete=False):
		
		if not self._started:
			self.startup()
		
		if complete:
			n = self.clock.get_remaining()
	
		out = self._take_steps(n)
		
		if self.clock.get_remaining() <= 0:
			print('Training Complete')
			ticks = self.clock.get_time()
			if self.records is not None and not self.records.check(ticks, self):
				self.records.activate(ticks, info=self)
			if self.vizualizer is not None and not self.vizualizer.check(ticks, self):
				self.vizualizer.activate(ticks, info=self)
			if self.validation is not None and not self.validation.check(ticks, self):
				self.validation.activate(ticks, info=self)
			if self.checkpointer is not None and not self.checkpointer.check(ticks, self):
				self.checkpointer.activate(ticks, info=self)
		
		return out
	
	def _take_steps(self, n=1):
		for i in range(n):
			self._take_step()
			
	def _take_step(self):
		self.clock.tick(self)
	
	def receive_batch(self, batch):
		self.out = None
		self.batch = batch
	
	def receive_output(self, batch, out):
		self.out = out
		
		records = self.get_records()
		records['total_steps']['train'] += 1
		records['total_samples']['train'] += batch.size(0)
	
	def startup(self):
		
		self._started = True
		
		A = self.get_config()
		clock = self.get_clock()

		dataset = self.get_dataset()
		clock.register_alert('data', dataset)
		model = self.get_model()
		clock.register_alert('train', model)
		
		clock.sort_alerts(start_with=['data', 'train'])
		
		records = self.get_records()
		clock.register_alert('log', records)
		
		if 'viz' not in A:
			A.push('viz._type', 'run/viz', overwrite=False)
		self.vizualizer = A.pull('viz', None)
		if self.vizualizer is not None:
			clock.register_alert('viz', self.vizualizer)
		
		self.validation = A.pull('validation', None)
		if self.validation is None:
			print('No validation')
		else:
			clock.register_alert('val', self.validation)
		
		if self.invisible:
			print('No checkpointing')
		else:
			A.push('checkpoint._type', 'run/checkpoint', overwrite=False, silent=True)
			self.checkpointer = A.pull('checkpoint', None)
		
			if self.checkpointer is None:
				print('No checkpointer found')
			else:
				clock.register_alert('save', self.checkpointer)
		
		if 'print' not in A:
			A.push('print._type', 'run/print')
		print_step = A.pull('print', None)
		if print_step is not None:
			clock.register_alert('print', print_step)
		
		ckpt = A.pull('clock._load-ckpt', None, silent=self.silent)
		if ckpt is not None:
			ckpt = Path(ckpt)
			clock.load_checkpoint(ckpt)
		
		path = self.get_path()
		if path is not None:
			config_path = self.get_path() / 'config.yaml'
			if not config_path.is_file():
				A.export(config_path)
		
		mode = 'train'
		
		dataset.switch_to(mode)
		model.switch_to(mode)
		records.switch_to(mode)
		
		clock.prep(self)
		# dataset.prep(model=model, records=records)
		# model.prep(dataset=dataset, records=records)
		# records.prep(dataset=dataset, model=model)
		
		if mode not in records['total_steps']:
			records['total_steps'][mode] = 0
		if mode not in records['total_samples']:
			records['total_samples'][mode] = 0
		if mode not in records['total_epochs']:
			records['total_epochs'][mode] = 0
			
		# sync clock
		
		print(records['total_steps'][mode])
		clock.set_time(records['total_steps'][mode])
		clock.sort_alerts(end_with=['save', 'print'], strict=False)
		
		if clock.get_remaining() is None:
			steps = A.pull('training.steps', '<>budget', None)
			if steps is None:
				epochs = A.pull('training.epochs', '<>epochs', None)
				if epochs is None:
					raise Exception('Clock has no limit')
				steps = len(dataset.get_loader()) * epochs
				
			clock.set_limit(steps)
		
		# self.loader = None
		# self.get_loader(activate=True, mode=mode, infinite=True)
	
	
	def evaluate(self, ident='eval', config=None,
	             evaluation=None, store_batch=True, overwrite=False, path=None):
		
		if config is not None:
			self.get_config().sub('eval').update(config)
		config = self.get_config().sub('eval')
		
		ident = config.push('ident', ident, overwrite=False)
		mode = config.pull('mode', 'val')
		assert ident is not None, 'No ident specified'
		
		config.push('mode', mode, force_root=True, silent=True)
		
		if path is None:
			path = self.get_path()
		if path is not None:
			path = Path(path)
			print(f'Will save results to {str(path)}')
			
		dataset = self.get_dataset()
		model = self.get_model()
		records = self.get_records()
		
		dataset.switch_to(mode)
		model.switch_to(mode)
		records.switch_to(mode)
		
		N = self.get_clock().get_time()
		fmt = '{}/{}'.format(ident, '{}')
		if config is not None:
			fmt = config.pull('log-fmt', fmt)
			
		_records_info = records.get_fmt(), records.get_step()
		records.set_fmt(fmt)
		records.set_step(N)
		
		if 'evaluations' not in records:
			records['evaluations'] = {}
		if ident not in records['evaluations']:
			records['evaluations'][ident] = []
		if N in records['evaluations'][ident]:
			overwrite = overwrite if config is None else config.pull('overwrite', False)
			if not overwrite:
				print(f'Already evaluated "{ident}" after {N} steps, and overwrite is False')
				results = None
				if config is not None and config.pull('load-results', False):
					results = self.get_results(ident)
				return results
			print(f'Overwriting {ident} at step {N}')
		else:
			records['evaluations'][ident].append(N)
		
		store_batch = store_batch if config is None else config.pull('store_batch', store_batch)
		
		if evaluation is None and config is not None:
			evaluation = config.pull('evaluation', None)
		
		batch, output = None, None
		if evaluation is not None:
			evaluation.set_mode(mode)
			def step(_batch, step, _records=None):
				nonlocal output, batch
				out = evaluation.epoch_step(_batch, step)
				if output is None:
					try:
						model.visualize(out, records)
					except AttributeError:
						pass
					batch = _batch
					output = out
				return out
			
			evaluation.run_epoch(dataset=dataset, model=model, records=None, step_fn=step)
		
			self.batch = batch
			self.output = output
		
		try:
			eval_out = model.evaluate(self, config=config)
		except AttributeError:
			pass
		else:
			if output is None:
				output = eval_out
			elif eval_out is not None:
				output.update(eval_out)
		
		records.step(fmt=fmt)
		
		if store_batch and output is not None:
			output['batch'] = batch
		if path is not None:
			self.update_results(ident, {k:v for k,v in output.items()}, path=path)
			records.checkpoint(path)
		
		fmt, N = _records_info
		records.set_fmt(fmt)
		records.set_step(N)
		
		return output
		
		
	def exit_run(self, cause, code=None, checkpoint=True):
		
		cmsg = f' (code={code})'
		
		if not self.silent:
			print(f'Exiting due to {cause}{cmsg}')
			
		if checkpoint and self.checkpointer is not None:
			print('Saving checkpoint')
			self.checkpointer.checkpoint()
		
		if code is not None:
			sys.exit(code)
	
	# endregion

@fig.AutoModifier('inline')
class Inline(Run):
	def _prep(self, A):
		
		inline = A.pull('inline', False)
		if inline:
			A.push('pbar._type', 'progress-bar', silent=True, overwrite=False)
		self.pbar = A.pull('pbar', None)
		if self.pbar is not None:
			A.push('print', None)

		super()._prep(A)

	def _take_steps(self, n=1):
		
		if self.pbar is not None and n > 1:
			self.pbar.init_pbar(limit=n)
		
		out = super()._take_steps(n=n)
		
		if self.pbar is not None:
			self.pbar.reset()
			
		return out
	
	def _take_step(self):
		
		super()._take_step()
		
		if self.pbar is not None:
			self.pbar.update(desc=self.get_inline_description())
	
	def get_inline_description(self):
		
		progress = self.model.get_description()
		ticks = self.clock.get_time()
		limit = self.clock.get_limit()
		
		mode = 'train'
		epochs = self.records['total_epochs'][mode]
		
		return f'{mode}:{epochs} {ticks}/{limit} {progress}'
		
		
# @fig.AutoModifier('extendable')
# class Extendable(Run):
#
# 	def startup(self, A):
# 		super().startup()
#
# 		A = self.get_config()
#
# 		extend = A.pull('extend', None)
# 		if extend is not None:
# 			self.clock.set_time(extend)


		
@fig.AutoModifier('timed-run')
class Timed(Run):
	
	def startup(self):
		super().startup()
		
		A = self.get_config()

		time_limit = A.pull('time_limit', None)
		if time_limit is not None:
			if not self.silent:
				print(f'SOFT TIME LIMIT: {time_limit:2.2f} hrs')
			self.timer_start = time.time()
			time_limit = 3600 * time_limit
		
			self.timer_exit_code = A.pull('time_limit_exit_code', 3)
		
		self.timer_limit = time_limit

	def save_checkpoint(self, save_model=True):
		super().save_checkpoint(save_model=save_model)
		
		time_limit = self.timer_limit
		if time_limit is not None:
			start_time = self.timer_start
			if (time.time() - start_time) > time_limit:
				self.exit_run('timelimit reached', code=self.timer_exit_code)
			

		



# from toposort import toposort as toposort_lib, toposort_flatten



# def toposort(root, get_edges, depth_first=False):
#
# 	order = [root]
# 	done = set(order)
# 	options = deque(get_edges(root))
#
# 	while len(options):
#
# 		next = None
# 		for node in options:
# 			success = True
# 			for check in options:
# 				if node in get_edges(check):
# 					success = False
# 					break
# 			if success:
# 				next = node
# 				break
# 		if next is None:
# 			raise CycleDetectedError(node)
# 		else:
# 			options.remove(next)
# 			if next not in done:
# 				order.append(next)
# 				done.add(next)
# 				if depth_first:
# 					options.extendleft(reversed(get_edges(next)))
# 				else:
# 					options.extend(get_edges(next))
#
# 	return order





from .layers import make_MLP, batched_grouped_linear, Reshaper, Recurrence, \
	FourierLayer, DenseLayer, ConvLayer, Interpolate, LayerNorm, Invertible, \
	ExpandedConvLayer, CoordConvLayer
from .nets import MLP, Multihead, MultiLayer
from .criterion import MultiGroupClassification, Feature_Match
# from .prior import PriorTfm, StorablePrior, AdaIN, StyleModel
from .features import Prior, Gaussian, Uniform, Normal
from .style import StyleFusionLayer, StyleFusion, PriorStyleFusionLayer, StyleExtractorLayer, StyleExtractor
from . import curriculum
from .unsup import Autoencoder, Variational_Autoencoder, Wasserstein_Autoencoder



import math
import torch

import omnifig as fig

from ..op import framework as fm
from ..op.clock import Freq, Reg, Savable

from .. import util


class Scheduler(util.Value, Savable, Reg, Freq):
	def __init__(self, A, **kwargs):
		super().__init__(A, **kwargs)
		self.start = A.pull('start', 0)
		self.stop = A.pull('stop', None)
		
		self.end = A.pull('end', '<>end', None)

		self.initial = self.get()
	
	def extra_repr(self):
		return {'end': self.end}
	
	def __repr__(self):
		extra = [f'    {k}: {v},' for k, v in self.extra_repr().items() if v is not None]
		
		return '{}[{}:{}:{}]({})'.format(self.__class__.__name__, '' if self.start is None else self.start,
		                                 '' if self.stop is None else self.stop, '' if self.freq is None else self.freq,
		                                 ('\n    ' + '\n'.join(extra) + '\n') if len(extra) else '')
	
	def __str__(self):
		return repr(self)
	
	def check(self, tick, info=None):
		return super().check(tick, info=info) \
		       and (self.start is None or tick >= self.start) \
		       and (self.stop is None or tick < self.stop)
	
	def activate(self, tick, info=None):
		x = self.step(self.get(), self.initial, tick, info=info)
		self.set(x)
		return x
	
	def state_dict(self):
		return {'end': self.end, 'start': self.start, 'stop': self.stop, 'val':self.get()}
	
	def load_state_dict(self, data):
		self.set(data['val'])
		if data is not None:
			for k, v in data.items():
				if k != 'val':
					setattr(self, k, v)
	
	def step(self, x, x0, tick, info=None):
		raise NotImplementedError


@fig.Component('scheduler/step')
class StepScheduler(Scheduler):
	def __init__(self, A, **kwargs):
		super().__init__(A, **kwargs)
		
		self.factor = A.pull('factor', '<>gamma')
	
	def state_dict(self):
		data = super().state_dict()
		data['factor'] = self.factor
		return data
	
	def step(self, x, x0, tick, info=None):
		return self.factor * x


@fig.Component('scheduler/multi-step')
class MultiStepScheduler(StepScheduler):
	def __init__(self, A, **kwargs):
		super().__init__(A, **kwargs)
		
		self.freq = None
		
		num = A.pull('num-steps', 1)
		limit = self.stop
		if limit is None:
			limit = A.pull('limit')
		
		size = limit // num
		self.steps = set([size * step for step in range(1, num)])
	
	def state_dict(self):
		data = super().state_dict()
		data['steps'] = self.steps
		return data
	
	def check(self, tick, info=None):
		return super().check(tick, info=info) and tick in self.steps


class FunctionScheduler(Scheduler):
	def __init__(self, A, **kwargs):
		super().__init__(A, **kwargs)
		
		assert self.end is not None, 'must have a min for exponential scheduler'
		
		self.limit = A.pull('limit', '<>stop')
		
		self.skew = math.exp(A.pull('skew', 0)) # +/- -> left/right
		self.kurtosis = 3. ** A.pull('kurtosis', 0) # +/- -> squeeze/stretch middle
		self.constrain_progress = A.pull('constrain-progress', False)
	
	def state_dict(self):
		data = super().state_dict()
		data.update({'limit': self.limit, 'skew': self.skew, 'kurtosis': self.kurtosis,
		             'constrain_progress': self.constrain_progress})
		return data
	
	def step(self, x, x0, tick, info=None):
		progress = (tick / self.limit) #** self.skew
		progress = 2*progress - 1
		progress = progress**self.kurtosis if progress >= 0 else -(-progress)**self.kurtosis
		progress = 0.5*progress + 0.5
		progress **= self.skew
		if self.constrain_progress:
			progress = max(0., min(progress, 1.))
		val = self._func(progress, x0)
		return val
	
	def _func(self, progress, x0):
		raise NotImplementedError


@fig.Component('scheduler/cos')
class CosScheduler(FunctionScheduler):
	def __init__(self, A, **kwargs):
		trig = A.pull('trig', 'cos')
		nodes = A.pull('half-periods', 1)
		super().__init__(A, **kwargs)
		if trig == 'cos':
			self.trig = math.cos
		elif trig == 'sin':
			self.trig = math.sin
		else:
			raise ValueError(trig)
		self.nodes = nodes
		
	def _func(self, progress, x0):
		return (x0 - self.end) * (self.trig(self.nodes * math.pi * progress) + 1) / 2 + self.end


@fig.Component('scheduler/exp')
class ExpScheduler(FunctionScheduler):
	def _func(self, progress, x0):
		return x0 * (self.end / x0) ** progress


@fig.Component('scheduler/lin')
class LinScheduler(FunctionScheduler):
	def _func(self, progress, x0):
		return x0 - (x0 - self.end) * progress


#### old



# class Scheduler(Freq):
# 	def __init__(self, A, **kwargs):
# 		super().__init__(A, **kwargs)
# 		self.start = A.pull('start', 0)
# 		self.stop = A.pull('stop', None)
#
# 		self.min_lr = A.pull('min_lr', None)
#
# 		self.optimizer = None
#
# 	def extra_repr(self):
# 		return {'min': self.min_lr}
#
# 	def __repr__(self):
# 		extra = [f'    {k}: {v},' for k,v in self.extra_repr().items() if v is not None]
#
# 		return '{}[{}:{}:{}]({})'.format(self.__class__.__name__, '' if self.start is None else self.start,
# 		                             '' if self.stop is None else self.stop, '' if self.freq is None else self.freq,
# 		                                 ('\n    '+'\n'.join(extra)+'\n') if len(extra) else '')
#
# 	def prep(self, optim):
# 		self.optimizer = optim
# 		self._initial_lrs = [param_group['lr'] for param_group in self.optimizer.param_groups]
#
# 	def check(self, tick, info=None):
# 		return super().check(tick, info=info) \
# 		        and (self.start is None or tick >= self.start) \
# 		        and (self.stop is None or tick < self.stop)
#
# 	def activate(self, tick, info=None):
# 		lr = None
# 		for param_group, lr0 in zip(self.optimizer.param_groups, self._initial_lrs):
# 			lr = param_group['lr']
# 			lr = self.compute_lr(lr, lr0, tick, info=info)
# 			if self.min_lr is not None:
# 				lr = max(self.min_lr, lr)
# 			param_group['lr'] = lr
#
# 		return lr
#
# 	def state_dict(self):
# 		return {'min_lr':self.min_lr, 'start':self.start, 'stop':self.stop}
#
# 	def load_state_dict(self, data):
# 		if data is not None:
# 			for k, v in data.items():
# 				setattr(self, k, v)
#
# 	def compute_lr(self, lr, lr0, tick, info=None):
# 		raise NotImplementedError
#
#
# @fig.Component('scheduler/step')
# class StepScheduler(Scheduler):
# 	def __init__(self, A, **kwargs):
# 		super().__init__(A, **kwargs)
#
# 		self.factor = A.pull('factor', '<>gamma')
#
# 	def state_dict(self):
# 		data = super().state_dict()
# 		data['factor'] = self.factor
# 		return data
#
# 	def compute_lr(self, lr, lr0, tick, info=None):
# 		return self.factor * lr
#
#
# @fig.Component('scheduler/multi-step')
# class MultiStepScheduler(StepScheduler):
# 	def __init__(self, A, **kwargs):
#
# 		super().__init__(A, **kwargs)
#
# 		self.freq = None
#
# 		num = A.pull('num-steps', 1)
# 		limit = self.stop
# 		if limit is None:
# 			limit = A.pull('limit')
#
# 		size = limit // num
# 		self.steps = set([size*step for step in range(1,num)])
#
# 	def state_dict(self):
# 		data = super().state_dict()
# 		data['steps'] = self.steps
# 		return data
#
# 	def check(self, tick, info=None):
# 		return super().check(tick, info=info) and tick in self.steps
#
#
# class FunctionScheduler(Scheduler):
# 	def __init__(self, A, **kwargs):
# 		super().__init__(A, **kwargs)
#
# 		assert self.min_lr is not None, 'must have a min for exponential scheduler'
#
# 		self.limit = A.pull('limit', '<>stop')
#
# 		self.power = A.pull('power', 1)
# 		self.constrain_progress = A.pull('constrain-progress', False)
#
#
# 	def state_dict(self):
# 		data = super().state_dict()
# 		data.update({'limit':self.limit, 'power':self.power,
# 		             'constrain_progress':self.constrain_progress})
# 		return data
#
# 	def compute_lr(self, lr, lr0, tick, info=None):
# 		progress = (tick/self.limit) ** self.power
# 		if self.constrain_progress:
# 			progress = max(0.,min(progress,1.))
# 		return self._func(progress, lr0)
#
# 	def _func(self, progress, lr0):
# 		raise NotImplementedError
#
#
# @fig.Component('scheduler/cos')
# class CosScheduler(FunctionScheduler):
# 	def _func(self, progress, lr0):
# 		return (lr0-self.min_lr) * (math.cos(math.pi * progress) + 1) / 2 + self.min_lr
#
#
# @fig.Component('scheduler/exp')
# class ExpScheduler(FunctionScheduler):
# 	def _func(self, progress, lr0):
# 		return lr0 * (self.min_lr/lr0) ** progress
#
#
# @fig.Component('scheduler/lin')
# class LinScheduler(FunctionScheduler):
# 	def _func(self, progress, lr0):
# 		return lr0 - (lr0-self.min_lr) * progress




## region old

class RunningNormalization(fm.FunctionBase):
	def __init__(self, dim, cmin=-5, cmax=5):
		super().__init__(dim, dim)
		self.dim = dim
		self.n = 0
		self.cmin, self.cmax = cmin, cmax

		self.register_buffer('sum_sq', torch.zeros(dim))
		self.register_buffer('sum', torch.zeros(dim))
		self.register_buffer('mu', torch.zeros(dim))
		self.register_buffer('sigma', torch.ones(dim))

	def update(self, xs):
		xs = xs.view(-1, self.dim)
		self.n += xs.shape[0]
		self.sum += xs.sum(0)
		self.mu = self.sum / self.n

		self.sum_sq += xs.pow(2).sum(0)
		self.mean_sum_sq = self.sum_sq / self.n

		if self.n > 1:
			self.sigma = (self.mean_sum_sq - self.mu**2).sqrt()

	def forward(self, x):
		if self.training:
			self.update(x)
		return ((x - self.mu) / self.sigma).clamp(self.cmin, self.cmax)

# endregion



import numpy as np
import torch
import torch.distributions as distrib
# from sklearn.decomposition import PCA
# import matplotlib.pyplot as plt

import omnifig as fig

# from ... import framework as fm
from ...op import framework as fm
from ... import util

from ..features import Prior, Gaussian


@fig.Component('ae')
class Autoencoder(fm.Regularizable, fm.Encodable, fm.Decodable, fm.Model):
	def __init__(self, A, encoder=None, decoder=None, din=None, dout=None, **other):

		if encoder is None:
			encoder = A.pull('encoder')
		if decoder is None:
			decoder = A.pull('decoder')

		reg_wt = A.pull('reg-wt', None)
		reg = A.pull('reg', 'L2' if reg_wt is not None and reg_wt > 0 else None)

		criterion = A.pull('criterion', 'mse')

		viz_latent = A.pull('viz-latent', True)
		viz_rec = A.pull('viz-rec', True)
		viz_enc = A.pull('viz-enc', True)
		viz_dec = A.pull('viz-dec', True)

		super().__init__(A, din=encoder.din, dout=decoder.dout, **other)
		
		self.encoder = encoder
		self.decoder = decoder

		self.criterion = util.get_loss_type(criterion)

		self.reg_wt = reg_wt
		self.reg_fn = util.get_regularization(reg, reduction='sum')

		if self.reg_wt is not None and self.reg_wt > 0:
			self.register_stats('reg-loss')
			self.register_hparam('reg_wt', reg_wt)
		self.register_stats('rec-loss')

		self.latent_dim = self.decoder.din

		self._viz_settings = set()
		if viz_latent:
			self._viz_settings.add('latent')
		if viz_rec:
			self._viz_settings.add('rec')
		if viz_enc:
			self._viz_settings.add('enc')
		if viz_dec:
			self._viz_settings.add('dec')

	def _visualize(self, info, records):
		
		settings = self._viz_settings
		
		if 'enc' in settings and isinstance(self.encoder, fm.Visualizable):
			self.encoder.visualize(info, records)
		if 'dec' in settings and isinstance(self.decoder, fm.Visualizable):
			self.decoder.visualize(info, records)
		
		if 'latent' in info:
			q = info.latent
			if 'latent' in settings and q is not None:
				if isinstance(info.latent, distrib.Distribution):
					q = q.loc

				shape = q.size()
				if len(shape) > 1 and np.product(shape) > 0:
					try:
						records.log('histogram', 'latent-norm', q.norm(p=2, dim=-1))
						records.log('histogram', 'latent-std', q.std(dim=0))
						if isinstance(info.latent, distrib.Distribution):
							records.log('histogram', 'logstd-hist', info.latent.scale.add(1e-5).log().mean(dim=0))
					except ValueError:
						print('\n\n\nWARNING: histogram just failed\n')
						print(q.shape, q.norm(p=2, dim=-1).shape)

		if 'original' in info:
			X = info.original
			if X.ndim == 4 and 'rec' in settings and 'reconstruction' in info:
				B, C, H, W = info.original.shape
				N = min(B, 8)

				R = info.reconstruction
				viz_x, viz_rec = X[:N], R[:N]

				recs = torch.cat([viz_x, viz_rec], 0)
				records.log('images', 'rec', util.image_size_limiter(recs))

	def forward(self, x, ret_q=False):

		q = self.encode(x)
		x = self.decode(q)

		if ret_q:
			return x, q
		return x

	def encode(self, x, **kwargs):
		return self.encoder(x, **kwargs)

	def regularize(self, q, p=None):
		B = q.size(0)
		mag = self.reg_fn(q)
		return mag / B

	def decode(self, q, **kwargs):
		return self.decoder(q, **kwargs)

	def preprocess(self, x):
		return x

	def _rec_step(self, out):

		x = out.original

		B = x.size(0)

		if 'reconstruction' not in out:

			rec, q = self(x, ret_q=True)

			if 'latent' not in out:
				out.latent = q
			out.reconstruction = rec

		rec = out.reconstruction

		loss = self.criterion(rec, x) / B
		out.rec_loss = loss
		self.mete('rec-loss', loss)

		return loss

	def _reg_step(self, out):

		if 'latent' not in out:
			out.latent = self.encode(out.original)

		q = out.latent

		reg_loss = self.regularize(q)
		self.mete('reg-loss', reg_loss)
		out.reg_loss = reg_loss
		return self.reg_wt * reg_loss

	def _process_batch(self, batch, out=None):
		if out is None:
			out = util.TensorDict()

		if isinstance(batch, torch.Tensor):
			x = batch
		elif isinstance(batch, (tuple, list)):
			x = batch[0]
		elif isinstance(batch, dict):
			x = batch['x']
		else:
			raise NotImplementedError

		x = self.preprocess(x)

		out.original = x
		out.batch = batch

		return out

	def _step(self, batch, out=None):

		out = self._process_batch(batch, out)

		if self.train_me():
			self.optim.zero_grad()

		loss = self._rec_step(out)

		if self.reg_wt is not None and self.reg_wt > 0:
			loss += self._reg_step(out)

		out.loss = loss

		if self.train_me():
			loss.backward()
			self.optim.step()

		return out

class Generative_AE(fm.Generative, Prior, Autoencoder):

	def __init__(self, A, **kwargs):
		super().__init__(A, **kwargs)
		self.prior_dim = self.latent_dim

	def generate(self, N=1, prior=None):
		if prior is None:
			prior = self.sample_prior(N)
		return self.decode(prior)


@fig.Component('vae')
class Variational_Autoencoder(Gaussian, Generative_AE):

	def __init__(self, A, **kwargs):

		# mod_check = A.pull('mod_check', True)  # make sure encoder outputs a normal distribution
		# if mod_check:
		# 	mods = A.pull('encoder._mod', None, silent=True)
		# 	if isinstance(mods, (list, tuple, dict)):
		# 		if 'normal' not in mods:
		# 			mods = [*mods, 'normal'] if isinstance(mods, (tuple, list)) else {**mods, 'normal': 10},
		# 			A.push('encoder._mod', mods, silent=True)
		# 	else:
		# 		A.push('encoder._mod.normal', 1)

		A.push('reg', None)  # already taken care of
		wt = A.pull('reg-wt', None, silent=True)
		if wt is None or wt <= 0:
			print('WARNING: VAEs must have a reg_wt (beta), setting to 1')
			A.push('reg-wt', 1)

		super().__init__(A, **kwargs)

	def decode(self, q):
		if isinstance(q, distrib.Distribution):
			q = q.rsample() if self.training else q.mean
		return super().decode(q)

	def regularize(self, q):
		return util.standard_kl(q).sum() / q.loc.size(0)


@fig.Component('wae')
class Wasserstein_Autoencoder(Generative_AE):  # MMD
	def __init__(self, A, **kwargs):
		A.push('reg', None)  # already taken care of
		super().__init__(A, **kwargs)

	def regularize(self, q, p=None):
		if p is None:
			p = self.sample_prior(q.size(0))
		return util.MMD(p, q).sum() / q.size(0)







import sys, os

import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision

import omnifig as fig

import numpy as np
# %matplotlib tk
from matplotlib import cm

import omnilearn as learn
from omnilearn import util


@fig.Component('point-enc')
class PointEncoder(learn.Encodable, learn.Visualizable, learn.FunctionBase):

	def __init__(self, A):
		in_shape = A.pull('in_shape', '<>din')
		latent_dim = A.pull('latent_dim', '<>dout')

		A.push('in_shape', in_shape, overwrite=False, silent=True)

		transform = A.pull('transform')  # converts the image into a point cloud
		pout, N = transform.dout


		A.push('pin', pout)
		A.push('n_points', N)

		pointnet = A.pull('pointnet')

		super().__init__(in_shape, latent_dim)

		self.transform = transform
		self.pointnet = pointnet

	def visualize(self, info, logger):
		if not self.training or self._viz_counter % 2 == 0:
			super().visualize(info, logger)
		else:
			self._viz_counter += 1

	def _visualize(self, out, logger):  # TODO
		for m in self.pointnet.tfms:
			if isinstance(m, learn.Visualizable):
				m._visualize(out, logger)

	def forward(self, x):  # images

		p = self.transform(x)
		q = self.pointnet(p)

		return q


@fig.AutoComponent('patch-points')
class Patch_Points(learn.FunctionBase):
	'''
	Converts an image into a set of unordered points where each point is a concat of the pixels of
	a unique patch of the image.
	'''

	def __init__(self, in_shape, include_coords=False,
	             patch_size=4, stride=2, dilation=1, padding=0):

		C, H, W = in_shape

		try:
			len(patch_size)
		except TypeError:
			patch_size = (patch_size, patch_size)

		try:
			len(stride)
		except TypeError:
			stride = (stride, stride)

		try:
			len(dilation)
		except TypeError:
			dilation = (dilation, dilation)

		try:
			len(padding)
		except TypeError:
			padding = (padding, padding)

		pout = C * patch_size[0] * patch_size[1]

		Nh = (H + 2 * padding[0] - dilation[0] * (patch_size[0] - 1) - 1) // stride[0] + 1
		Nw = (W + 2 * padding[1] - dilation[1] * (patch_size[1] - 1) - 1) // stride[1] + 1
		N = Nh * Nw

		if include_coords:
			pout += 2

		super().__init__(in_shape, (pout, N))

		self.pout = pout

		self.transform = nn.Unfold(kernel_size=patch_size, stride=stride,
		                           dilation=dilation, padding=padding)

		if include_coords:
			self.register_buffer('coords', torch.from_numpy(np.mgrid[:1:1j * Nh, :1:1j * Nw]).view(2, -1).float())
		else:
			self.coords = None

	def forward(self, x):
		p = self.transform(x)
		if self.coords is not None:
			B = p.size(0)
			p = torch.cat([p, self.coords.expand(B, *self.coords.size())], 1)
		return p


@fig.AutoComponent('1dconv-net')
def make_pointnet(pin, pout, hidden=None,
                  nonlin='prelu', output_nonlin=None):
	if hidden is None:
		hidden = []
	nonlins = [nonlin] * len(hidden) + [output_nonlin]
	hidden = [pin] + list(hidden) + [pout]

	layers = []
	for in_dim, out_dim, nonlin in zip(hidden, hidden[1:], nonlins):
		layers.append(nn.Conv1d(in_dim, out_dim, kernel_size=1))
		if nonlin is not None:
			layers.append(util.get_nonlinearity(nonlin))

	net = nn.Sequential(*layers)
	net.pin = pin
	net.pout = pout
	return net


@fig.Component('point-net')
class PointNet(learn.Model):
	def __init__(self, A):

		pin = A.pull('pin')  # should be the number of channels of the points
		dout = A.pull('latent_dim', '<>dout')

		n_points = A.pull('n_points', None)

		create_module = A.pull('modules')

		super().__init__(din=pin, dout=dout)

		modules = []

		nxt = create_module.view()
		# assert nxt is not None, 'no point-net modules provided'
		nxt.push('pin', pin, silent=True)
		nxt.push('N', n_points, silent=True)

		for module in create_module:

			if hasattr(module, 'nout'):
				n_points = module.nout

			modules.append(module)
			try:
				nxt = create_module.view()
			except StopIteration:
				break
			else:
				nxt.push('pin', module.pout, silent=True)
				nxt.push('pin1', module.pout1, silent=True)
				nxt.push('pin2', module.pout2, silent=True)
				nxt.push('N', n_points, silent=True)

				if nxt.pull('_type', None, silent=True) == 'point-dual':
					if 'left' in nxt:
						nxt.push('left.pin', module.pout1, silent=True)
					if 'right' in nxt:
						nxt.push('right.pin', module.pout2, silent=True)

		pout = module.pout
		assert pout is not None, f'last module must have a single output: {module}'

		self.tfms = nn.Sequential(*modules)

		pool_type = A.pull('pool._type', None, silent=True)
		if pool_type is not None:
			A.push('pool.din', (pout, n_points), silent=True)
		# A.pool.pin = pout
		# A.pool.N = n_points
		self.pool = A.pull('pool', None)
		if self.pool is not None:
			pout = self.pool.dout

		final_type = A.pull('final._type', None, silent=True)
		if final_type is not None:
			A.push('final.din', pout, silent=True)
			A.push('final.dout', dout, silent=True)
		final = A.pull('final', None)

		if final is not None:  # make sure final will produce the right output
			assert final.din == pout and final.dout == dout, f'invalid {final.din} v {pout} and {final.dout} v {dout}'
		elif pout != dout:  # by default fix output to correctly sized output using a dense layer
			final = nn.Linear(pout, dout)
		self.final = final

	def forward(self, pts):

		pts = self.tfms(pts)

		B, _, *rest = pts.size()
		if len(rest):
			if rest[0] > 1:
				pts = self.pool(pts)
			pts = pts.view(B, -1)
		out = pts

		if self.final is not None:
			out = self.final(out)

		return out


# Abstract modules

class PointNetModule(learn.FunctionBase):
	def __init__(self, pin=None, pout=None,
	             pin1=None, pout1=None, pin2=None, pout2=None):
		super().__init__(pin, pout)
		self.pin = pin
		self.pout = pout

		self.pin1 = pin1
		self.pout1 = pout1
		self.pin2 = pin2
		self.pout2 = pout2


class PointSplit(PointNetModule):
	def __init__(self, pin, pout1, pout2):
		super().__init__(pin=pin, pout1=pout1, pout2=pout2)


class PointTransform(PointNetModule):
	def __init__(self, pin, pout):
		super().__init__(pin=pin, pout=pout)


class PointParallel(PointNetModule):
	def __init__(self, pin1, pin2, pout1, pout2):
		super().__init__(pin1=pin1, pin2=pin2, pout1=pout1, pout2=pout2)

	def __call__(self, pts):  # so the point net can connect modules by nn.Sequential
		return super().__call__(*pts)


class PointJoin(PointNetModule):
	def __init__(self, pin1, pin2, pout):
		super().__init__(pin1=pin1, pin2=pin2, pout=pout)

	def __call__(self, pts):  # so the point net can connect modules by nn.Sequential
		return super().__call__(*pts)


# Point-net operations


@fig.AutoComponent('point-split')
class PointSplitter(PointSplit):
	def __init__(self, pin, split):
		super().__init__(pin=pin, pout1=split, pout2=pin - split)

		assert pin > split, f'not enough dimensions to split: {pin} vs {split}'
		self.split = split

	def extra_repr(self):
		return f'split={self.split}'

	def forward(self, p):
		return p[:, :self.split], p[:, self.split:]


@fig.AutoComponent('point-buffer')
class PointBuffer(PointSplit):
	def __init__(self, pin, channels, N):
		super().__init__(pin=pin, pout1=channels, pout2=pin)

		self.buffer = nn.Parameter(torch.randn(channels, N), requires_grad=True)

	def extra_repr(self):
		return 'chn={}, N={}'.format(*self.buffer.size())

	def forward(self, p):
		B = p.size(0)

		p1 = self.buffer.expand(B, *self.buffer.shape)
		p2 = p

		return p1, p2


@fig.AutoComponent('point-transform-net')
class PointTransformNet(PointTransform):
	def __init__(self, net):
		super().__init__(net.pin, net.pout)

		self.net = net

	def forward(self, p):
		return self.net(p)


@fig.AutoComponent('point-self')
class PointSelfTransform(PointTransformNet):
	def __init__(self, pin, pout, hidden=None, nonlin='prelu', output_nonlin=None):
		super().__init__(make_pointnet(pin, pout, hidden=hidden, nonlin=nonlin,
		                               output_nonlin=output_nonlin))


@fig.AutoComponent('point-dual')
class PointDualTransform(PointParallel):
	def __init__(self, left=None, right=None, pin1=None, pin2=None):

		assert left is not None or pin1 is not None
		assert right is not None or pin2 is not None

		if left is not None:
			pin1 = left.pin
			pout1 = left.pout
		else:
			pout1 = pin1

		if right is not None:
			pin2 = right.pin
			pout2 = right.pout
		else:
			pout2 = pin2

		super().__init__(pin1, pin2, pout1, pout2)

		self.left = left
		self.right = right

	def forward(self, p1, p2):
		if self.left is not None:
			p1 = self.left(p1)
		if self.right is not None:
			p2 = self.right(p2)
		return p1, p2


@fig.AutoComponent('point-swap')
class PointSwap(PointParallel):
	def __init__(self, pin1, pin2):
		super().__init__(pin1=pin1, pin2=pin2, pout1=pin2, pout2=pin1)

	def forward(self, p1, p2):
		return p2, p1


@fig.AutoComponent('point-cat')
class PointJoiner(PointJoin):
	def __init__(self, pin1, pin2):
		super().__init__(pin1=pin1, pin2=pin2, pout=pin1 + pin2)

	def forward(self, p1, p2):  # TODO make sure shapes work out
		return torch.cat([p1, p2], dim=1)


@fig.AutoComponent('point-wsum')
class PointWeightedSum(learn.Cacheable, learn.Visualizable, PointJoin):

	def __init__(self, pin1, pin2, heads=1, keys=1, norm_heads=False, sum_heads=True,
	             gumbel=None, gumbel_min=0.1, gumbel_delta=2e-4):
		super().__init__(pin1=pin1, pin2=pin2, pout=pin1)
		self.nout = heads if sum_heads else heads * keys

		self.weights = nn.Conv1d(pin2, heads * keys, kernel_size=1)

		self.norm_heads = norm_heads
		self.sum_heads = sum_heads

		self.keys = keys
		self.heads = heads
		self.N_out = self.keys * self.heads

		if gumbel is not None and gumbel <= 0:
			gumbel = None
		self.gumbel = gumbel
		self.gumbel_delta = gumbel_delta
		self.gumbel_start = gumbel
		self.gumbel_min = gumbel_min

		self.register_cache('_w')
		self.cmap = cm.get_cmap('seismic')

	def compute_weights(self, p):  # optionally use gumbel softmax
		return self.weights(p)

	def extra_repr(self):
		return f'heads={self.heads}, keys={self.keys}'

	def _visualize(self, out, logger):

		if self._w is not None:
			w = self._w

			brightness = 100 * 10 ** (self.gumbel is None)

			B, G, K, N = w.shape

			H, W = util.calc_tiling(N)

			g = w[0]
			g = g.view(G, K, H, W)
			out.key_selections = g
			g = g.view(G * K, H, W)

			g = torchvision.utils.make_grid(g.unsqueeze(1).mul(brightness).clamp(max=1), nrow=K, padding=1,
			                                pad_value=1)[:1]  # .mean(dim=0).unsqueeze(0)

			# g = torch.from_numpy(self.cmap(g.numpy())).permute(0, 3, 1, 2)[:,:3]
			# g = torchvision.utils.make_grid(g, nrow=K, padding=1, pad_value=1)#.norm(p=1, dim=0)

			logger.add('image', 'patches-1', g)

			g = w.sum(0) / B
			g = g.view(G, K, H, W)
			out.key_selections = g
			g = g.view(G * K, H, W)

			g = torchvision.utils.make_grid(g.unsqueeze(1).mul(brightness).clamp(max=1), nrow=K, padding=1,
			                                pad_value=1)[:1]  # .mean(dim=0).unsqueeze(0)

			# g = torch.from_numpy(self.cmap(g.numpy())).permute(0, 3, 1, 2)[:,:3]
			# g = torchvision.utils.make_grid(g, nrow=K, padding=1, pad_value=1)#.norm(p=1, dim=0)

			logger.add('image', 'patches-avg', g)

	def forward(self, p1, p2):  # p1 (B, C, N)
		w = self.compute_weights(p2)  # (B, GK, N)
		B, GK, N = w.shape
		G = self.heads
		K = self.keys

		if self.gumbel is not None:
			if self.training:
				self.gumbel = max(self.gumbel - self.gumbel_delta, self.gumbel_min)
			w += torch.rand_like(w).log().mul(-1).log().mul(-1)
			w /= self.gumbel

		w = w.view(B, G, K * N) if self.norm_heads else w.view(B, G, K, N)
		w = F.softmax(w, dim=-1).view(B, G * K, N)

		self._w = w.view(B, G, K, N).cpu().detach()

		w = w.permute(0, 2, 1)
		v = p1 @ w

		if self.sum_heads:
			C = p1.size(1)
			v = v.view(B, C, G, K).sum(-1)

		return v  # (B, C, G*K)


# @fd.AutoComponent('point-wsum')
# class PointWeightedSum(fd.Cacheable, fd.Visualizable, PointJoin):
#
# 	pass


@fig.AutoComponent('pool-points')
class Pool_Points(PointTransform):
	def __init__(self, pin, fn='max', p=2):
		super().__init__(pin, pin)

		assert fn in {'max', 'avg', 'sum', 'std', 'norm'}, f'unknown pool type: {fn}'

		self.fn = fn
		self.p = p

	def extra_repr(self):
		return f'{self.fn}'

	def forward(self, p):

		if self.fn == 'max':
			return p.max(-1, keepdim=True)[0]
		if self.fn == 'avg':
			return p.mean(-1, keepdim=True)
		if self.fn == 'sum':
			return p.sum(-1, keepdim=True)
		if self.fn == 'std':
			return p.std(-1, keepdim=True)
		if self.fn == 'norm':
			return p.norm(p=self.p, dim=-1, keepdim=True)


@fig.AutoComponent('concat-points')
class Concat_Points(PointTransform):
	def __init__(self, pin, N=None, groups=1):
		if N is None:
			print('WARNING: no number of points set')
			N = 1

		super().__init__(pin, N * pin // groups)

		self.groups = groups

	def forward(self, p):
		B = p.size(0)
		return p.permute(0, 2, 1).contiguous().view(B, self.groups, -1)








import sys, os
from pathlib import Path
import torch
import numpy as np
import cv2
import matplotlib.pyplot as plt

import requests

from omnibelt import get_printer

prt = get_printer(__name__)

OMNILEARN_DIR = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))

# _warn_savedir = True
# def get_save_dir(A=None):
#
# 	if A is not None:
# 		root = A.pull('saveroot', None)
# 		if root is not None:
# 			return Path(root)
#
# 	if 'OMNILEARN_SAVE_DIR' in os.environ:
# 		return Path(os.environ['OMNILEARN_SAVE_DIR'])
#
# 	root = Path(os.getcwd())
# 	root = root / 'trained_nets'
# 	root.mkdir(exist_ok=True)
#
# 	global _warn_savedir
# 	if _warn_savedir:
# 		prt.warning(f'No savedir found (specify with "OMNILEARN_SAVE_DIR" env variable), '
# 		            f'now using {str(root)}')
# 		_warn_savedir = False
#
# 	return root

_warn_datadir = True
def get_data_dir(A=None, silent=False):
	
	if A is not None:
		root = A.pull('dataroot', None, silent=silent)
		if root is not None:
			return Path(root)
		
	if 'OMNILEARN_DATA_DIR' in os.environ:
		return Path(os.environ['OMNILEARN_DATA_DIR'])
	
	root = Path(os.getcwd())
	root = root / 'local_data'
	root.mkdir(exist_ok=True)
	
	global _warn_datadir
	if _warn_datadir:
		prt.warning(f'No datadir found (specify with "OMNILEARN_DATA_DIR" env variable), '
					f'now using {str(root)}')
		_warn_datadir = False
	
	return root


def save_figure(name, fg=None, root=None, force=False, exts=None, **kwargs):
	
	if exts is None:
		exts = ['png']
	
	if fg is None:
		fg = plt.gcf()
	
	if root is not None or force:
		
		if root is None:
			path = name
		else:
			path = str(Path(root) / name)
		
		for ext in exts:
			fg.savefig(path + f'.{ext}', **kwargs)
		
		print(f'Figure {name} saved as {set(exts)}')
	


def get_patch(path, ht=None, wd=None, ret_bgr=False):

	img = cv2.imread(path)

	if not ret_bgr:
		img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB) # COLOR_BGR2HSV

	H, W = img.shape[0], img.shape[1]

	if ht is None:
		ht = H
	if wd is None:
		wd = W

	i = np.random.randint(H - ht) if H > ht else 0
	j = np.random.randint(W - wd) if W > wd else 0
	patch = img[i:i+ht, j:j+wd]
	if patch.shape[:2] != (ht, wd):
		patch = cv2.resize(patch, (ht, wd))
	return patch

def get_img(path, ht=None, wd=None, ret_bgr=False, pad=False):
	if ht is not None and wd is None:
		wd = ht

	img = cv2.imread(path)
	H, W = img.shape[:2]

	if not ret_bgr:
		img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)

	if ht is None or (ht, wd) == (H,W):
		return img
	elif pad and ht is not None and H < ht:
		pimg = np.zeros((ht,wd,3), dtype=np.uint8)
		pH, pW = (ht-H)//2, (wd-W)//2
		pimg[pH:pH+H, pW:pW+W] = img
		return pimg

	return cv2.resize(img, (wd, ht))



def download_file_from_google_drive(ID, out_dir, name=None, pbar=None, chunk_size=32768):
	URL = "https://docs.google.com/uc?export=download"

	session = requests.Session()

	response = session.get(URL, params = { 'id' : ID}, stream = True)
	token = get_confirm_token(response)

	if token is None:
		response.raise_for_status()

	params = { 'id' : ID, 'confirm' : token}
	response = session.get(URL, params = params, stream = True)

	out_dir = Path(out_dir)
	out_dir.mkdir(exist_ok=True)
	
	if name is None:
		name = response.headers.get('Content-Disposition', '"download').split('"')[1]
	
	idx = 1
	if name is None:
		name = 'download'
	_name = Path(name)
	base, ext = _name.stem, _name.suffix
	while (out_dir / name).exists():
		name = f'{base}_{idx}.{ext}' if len(ext) else f'{base}_{idx}'
		idx += 1
	
	dest = out_dir / name
	
	save_response_content(response, dest, chunk_size=chunk_size, pbar=pbar)
	
	return dest

def get_confirm_token(response):
	for key, value in response.cookies.items():
		if key.startswith('download_warning'):
			return value

	return None

def save_response_content(response, destination, chunk_size=32768, pbar=None):
	
	total_size_in_bytes = int(response.headers.get('content-length', 0))
	if pbar is not None:
		progress_bar = pbar(total=total_size_in_bytes, unit='iB', unit_scale=True)
		progress_bar.set_description(f'Downloading {destination.name}')

	with open(str(destination), "wb") as f:
		for chunk in response.iter_content(chunk_size):
			if chunk: # filter out keep-alive new chunks
				f.write(chunk)
				if pbar is not None:
					progress_bar.update(len(chunk))
	
	if pbar is not None:
		progress_bar.close()







from .misc import *
from .farming import *
from .os import *
from .stats import *
from .data import *
from .logic import *
from .math import *
from .optim import *
from .containers import *
from .logging import *
from .features import *

#from .setup import *
try:
	from .viz import *
except:
	print('WARNING: Failed to import visualization utilities - possibly due to matplotlib issues')
#from .setup import *




import sys, os
from pathlib import Path

import random

from tqdm import tqdm
try:
	import wget
except ImportError:
	print('WARNING: unable to import wget')
import shutil
import h5py as hf

import cv2
import numpy as np
import torch

from omnibelt import create_dir
import omnifig as fig

from ... import util

from ...data import Dataset, Deviced, Batchable, Image_Dataset

DATASET_URL = 'ftp://guest:GU.205dldo@ftp.softronics.ch/mvtec_anomaly_detection/mvtec_anomaly_detection.tar.xz'

def get_images(root, size=None, is_mask=False):
	raw = [util.read_raw_bytes(name, root=root) for name in os.listdir(root)]
	if size is None:
		return raw
	
	images = []
	for data in raw:
		image = util.str_to_byte_img(data) if is_mask else util.str_to_rgb(data)[..., ::-1]
		image = cv2.resize(image, (size,size))
		images.append(image)
	return images
	
@fig.Script('download-mvtec', description='Download and Format MVTec Anomaly Detection dataset')
def download_mvtec(A):
	raise NotImplementedError
	root = A.pull('root', os.environ.get('OMNILEARN_DATA_DIR', None))
	
	if root is None:
		raise Exception('Must specify a OMNILEARN_DATA_DIR as env variable (or pass in a `root`)')
	
	root = os.path.join(root, 'mvtec')
	
	if not os.path.isdir(root):
		create_dir(root)

	raw_name = A.pull('raw-name', 'mvtec_anomaly_detection.tar.xz', silent=True)
	
	size = A.pull('size', None)
	name = 'full' if size is None else f'size{size}'
	
	path = os.path.join(root, name)
	
	cleanup = A.pull('cleanup', False)
	cleanup_compressed = A.pull('cleanup-compressed', False)
	
	base = os.path.join(root, 'raw')
	if not os.path.isdir(base):
		rawpath = os.path.join(root, raw_name)
		if not os.path.isfile(rawpath):
			url = A.pull('url', DATASET_URL)
			print(f'Downloading MVTec AD dataset from {url}')
			
			rawpath = wget.download(url, out=root)
		
		create_dir(base)
		
		print(f'Extracting MVTec AD dataset from {rawpath} file')
		
		shutil.unpack_archive(rawpath, extract_dir=base)
		
		if cleanup_compressed:
			os.remove(rawpath)
	
	create_dir(path)
	
	contents = os.listdir(path)
	
	todo = [name for name in os.listdir(base)
	        if os.path.isdir(os.path.join(base, name))
	        and f'{name}.h5' not in contents]
	
	print(f'Formatting {len(todo)} object directories')
	
	pbar = A.pull('pbar', True, silent=True)
	if pbar:
		todo = tqdm(todo)
	
	for cat in todo:
		
		croot = os.path.join(base, cat)
		
		if pbar:
			todo.set_description(cat)
		
		with hf.File(os.path.join(path, f'{cat}.h5'), 'w') as f:
		
			mroot = os.path.join(croot, 'ground_truth')
		
			defects = os.listdir(mroot)
			f.attrs['defects'] = defects
		
			for defect in defects:
				f.create_dataset(f'mask_{defect}', data=get_images(os.path.join(mroot, defect),
				                                                   size=size, is_mask=True))
		
			f.create_dataset('train_good', data=get_images(os.path.join(croot, 'train', 'good'), size=size))
			
			mroot = os.path.join(croot, 'test')
			for test in os.listdir(mroot):
				f.create_dataset(f'test_{test}', data=get_images(os.path.join(mroot, test), size=size))
		
		if cleanup:
			print(f'removing: {cat}')
			# shutil.rmtree(croot)
	
	print('Formatted all categories')
	
	return path

@Dataset('mvtec')
class MVTec_Anomaly_Detection(Batchable):

	CATEGORIES = {'bottle', 'carpet', 'leather', 'screw', 'transistor', 'cable', 'grid',
        'metal_nut', 'tile', 'wood', 'capsule', 'hazelnut', 'pill','toothbrush', 'zipper'}
	GREYSCALES = {'screw', 'grid', 'zipper'}

	def __init__(self, A, **kwargs):
	
		dataroot = util.get_data_dir(A) / 'mvtec'

		mode = A.pull('mode', 'train')

		size = A.pull('size', None)
		
		dirname = 'full' if size is None else f'size{size}'
		droot = dataroot / dirname
		
		if not dataroot.is_dir() or not droot.is_dir():
			download = A.pull('download', False)
			if not download:
				raise Exception(f'This dataset ({dirname}) hasnt been downloaded and setup yet '
				                '(set the "download" argument to do so automatically)')
		
			print(f'Downloading/Formatting MVTec dataset: {dirname}')
			droot = fig.quick_run('download-mvtec', root=str(dataroot), size=size)
			droot = Path(droot)
		
		cat = A.pull('cat', 'random')
		if cat == 'random':
			cat = random.choice(list(self.CATEGORIES))

		ratio = A.pull('ratio', None)
		if ratio is not None:
			assert 0 < ratio <= 1, f'bad ratio: {ratio}'

		path = [c for c in droot.glob('*.h5') if cat == c.stem][0]
		
		include_class = A.pull('include-class', '<>include_class', True)
		include_mask = A.pull('include-masks', '<>include_masks', False)
		
		cut = A.pull('cut', 'train')
		assert cut in {None, 'normal', 'anomalies', 'all', 'train', 'test'}, f'unknown: {cut}'

		tfms = A.pull('transforms', None, silent=True)
		if tfms is not None:
			tfms = tfms.get(cat, None)

		augmenter = None
		augment_factor = 1
		if tfms is not None:
			augment_here = A.pull('augment_factor', None)

			if augment_here is not None:
				A.push('augmenter', tfms)
				A.push('augmenter._type', 'image-transform', overwrite=False)
				augmenter = A.pull('augmenter', None)
				augment_factor = augment_here
			else:
				A.push('augmentations', tfms, force_root=True)



		C = 1 if cat in self.GREYSCALES else 3
		din = (C, 1024, 1024) if size is None else (C, size, size)
		dout = din
		super().__init__(A, din=din, dout=dout, **kwargs)
		
		raw = hf.File(path, 'r')
		
		# print(raw.keys())

		uses = [key for key in raw.keys() if 'mask_' not in key]
		if cut == 'train':
			uses = [key for key in uses if 'train_' in key]
		elif cut == 'test':
			uses = [key for key in uses if 'test_' in key]
		elif cut == 'normal':
			uses = [key for key in uses if 'train_' in key or key == 'test_good']
		elif cut == 'anomalies':
			uses = [key for key in uses if 'test_' in key and key != 'test_good']

		C, H, W = din
		
		images = []
		masks = [] if include_mask else None
		labels = [] if include_class else None
		idents = []
		for key in uses:
			N = len(raw[key])
			if ratio is None:
				sel = ()
			elif mode == 'test':
				sel = slice(int(-N*(1-ratio)),None)
			else:
				sel = slice(int(N*ratio))

			imgs = raw[key][sel]
			# print(key, imgs.shape)
			images.extend(imgs)
			if labels is not None:
				labels.extend([1 if 'test_' in key else 0]*len(imgs))
			ident = '_'.join(key.split('_')[1:])
			idents.append(ident)
			if masks is not None:
				if 'test_' in key and ident != 'good':
					ident = f'mask_{ident}'
					masks.extend(raw[ident][sel])
				else:
					masks.extend([np.zeros((H,W))]*len(imgs))
		
		if masks is not None and not len(masks):
			masks = None
		if labels is not None:
			labels = torch.tensor(labels).long()
		
		if size is not None:
			# print([m.shape for m in images])
			images = torch.from_numpy(np.stack(images))
			images = images.unsqueeze(1) if C == 1 else images.permute(0,3,1,2)
			images = images.float().div(255)
			if masks is not None and len(masks):
				# print([m.shape for m in masks])
				masks = torch.from_numpy(np.stack(masks)).unsqueeze(1).bool()

		raw.close()
		# self.raw = raw

		self.transforms = tfms
		
		self.size = size
		
		self.images = images
		self.masks = masks
		self.labels = labels
		self.uses = uses
		self.idents = idents

		self.augmenter = augmenter
		self.augment_factor = augment_factor

	def get_transforms(self):
		return self.transforms
	
	def __len__(self):
		return len(self.images)#*self.augment_factor
	
	def __getitem__(self, item):

		# item %= len(self.images)
		
		if self.size is not None:
			out = [self.images[item]]

			if self.masks is not None:
				out.append(self.masks[item])

				if self.augmenter is not None:
					full = self.augmenter(torch.cat(out, 1))
					out = [full[:, :-1].contiguous(), full[:, -1:].contiguous()]

				if self.labels is not None:
					out.append(self.labels[item])
			elif self.labels is not None:
				out.append(self.labels[item])
			if self.augmenter is not None:
				out[0] = self.augmenter(out[0])

			return out
		
		raise NotImplementedError








import numpy as np
import torch
from torch import nn
from torch.nn import functional as F

from omnibelt import unspecified_argument

from bisect import bisect_right

from ... import util

from omnifig import AutoModifier, Modification

from ...data import Dataset, Deviced, DatasetBase, Batchable, Subset_Dataset

@Dataset('concat')
class Concat(DatasetBase):
	def __init__(self, A, din=None, dout=None, **kwargs):
		datasets = A.pull('datasets')
		assert len(datasets), 'no datasets'
		super().__init__(A, din=datasets[0].din, dout=datasets[0].dout, **kwargs)

		self.cumlens = np.cumsum([len(dataset) for dataset in datasets])
		self.datasets = datasets

	def __len__(self):
		return self.cumlens[-1]

	def __getitem__(self, item):
		idx = bisect_right(self.cumlens, item)
		return self.datasets[idx][item - int(self.cumlens[int(idx-1)]) if idx > 0 else item]

@AutoModifier('cropped')
class Cropped(DatasetBase):
	'''
	Parent dataset must have a din that is an image

	'''

	def __init__(self, A, crop_size=unspecified_argument, **kwargs):

		if crop_size is unspecified_argument:
			crop_size = A.pull('crop_size', None)

		crop_loc = A.pull('crop_loc', 'center')
		# crop_key = A.pull('crop_key', None) # TODO

		if crop_loc is not 'center':
			raise NotImplementedError

		if crop_size is not None:
			try:
				len(crop_size)
			except TypeError:
				crop_size = crop_size, crop_size
			assert len(crop_size) == 2, 'invalid cropping size: {}'.format(crop_size)
	
			if crop_size[0] != crop_size[1]:
				raise NotImplementedError('only square crops are implemented')
	
			assert hasattr(self, 'din'), 'This modifier requires a din (see Info_Dataset, eg. 3dshapes) ' \
			                                'in the dataset to be modified'
	
			assert len(self.din) == 3 or len(self.din) == 1, 'must be an image dataset'
	
			din = self.din
			
			self.din = (self.din[0], *crop_size)

		super().__init__(A, **kwargs)
		
		self.r = None
		if crop_size is not None:
		
			_, self.cy, self.cx = din
			self.cy, self.cx = self.cy // 2, self.cx // 2
			self.r = crop_size[0] // 2


	def __getitem__(self, item):
		sample = super().__getitem__(item)
		if self.r is None:
			return sample
		
		img, *other = sample

		img = img[..., self.cy-self.r:self.cy+self.r, self.cx-self.r:self.cx+self.r]

		return (img, *other)

# @AutoModifier('interpolated')
class Interpolated(DatasetBase):
	def __init__(self, A, interpolate_size=None, interpolate_mode=None, **kwargs):

		if interpolate_size is None:
			interpolate_size = A.pull('interpolate_size', None)

		if interpolate_mode:
			interpolate_mode = A.pull('interpolate_mode', 'bilinear')

		assert hasattr(self, 'din'), 'This modifier requires a din (see Info_Dataset, eg. 3dshapes) ' \
		                             'in the dataset to be modified'

		try:
			len(interpolate_size)
		except TypeError:
			interpolate_size = interpolate_size, interpolate_size
		assert len(interpolate_size) == 2, 'invalid cropping size: {}'.format(interpolate_size)

		assert len(self.din) == 3 or len(self.din) == 1, 'must be an image dataset'

		raise NotImplementedError
		A.din = (self.din[0], *interpolate_size)

		super().__init__(A, **kwargs)

		self.interpolate_size = interpolate_size
		self.interpolate_mode = interpolate_mode

	def __getitem__(self, item):

		sample = self.__getitem__(item)
		img, *other = sample

		img = F.interpolate(img, self.interpolate_size, mode=self.interpolate_mode).squeeze(0)

		return (img, *other)


@AutoModifier('resamplable')
class Resamplable(DatasetBase):
	def __init__(self, A, budget=None):
		if budget is None:
			budget = A.pull('budget', None)

		super().__init__(A)

		self.budget = budget
		self.inds = self.resample(self.budget)

	def resample(self, budget=None):
		if budget is None:
			budget = self.budget
		inds = None
		if budget is not None:
			inds = torch.randint(0, super().__len__(), size=(budget,))
		return inds

	def __len__(self):
		return super().__len__() if self.budget is None else self.budget

	def pre_epoch(self, mode, epoch):
		if self.budget is not None and mode == 'train':
			self.inds = self.resample()
		super().pre_epoch(mode, epoch)

	def __getitem__(self, item):
		if self.budget is not None:
			item = self.inds[item]
		return super().__getitem__(item)



@Modification('blurred')
def blurred(dataset, config):

	if not isinstance(dataset, Deviced):
		raise NotImplementedError

	level = config.pull('blur-level', 5)
	assert level % 2 == 1, f'bad blur level: {level}'
	blur_type = config.pull('blur-type', 'uniform')

	assert blur_type == 'uniform', f'not implemented: {blur_type}'

	pad_type = config.pull('pad-type', 'reflect')
	padding = (level-1)//2

	C = dataset.din[0]

	blur = nn.Conv2d(C, 1, groups=C, bias=False, kernel_size=level,
	                      padding=padding, padding_mode=pad_type)
	blur.weight.requires_grad = False
	blur.weight.copy_(torch.ones_like(blur.weight)).div_(level**2)

	blur.to(dataset.device)

	key = config.pull('data_key', 'images')
	full = getattr(dataset, key)

	with torch.no_grad():
		full = blur(full).detach()

	setattr(dataset, key, full)

	return dataset


@Modification('subset')
def make_subset(dataset, info):
	num = info.pull('num', None)
	
	shuffle = info.pull('shuffle', True)
	
	if num is None or num == len(dataset):
		print('WARNING: no subset provided, using original dataset')
		return dataset
	
	assert num <= len(dataset), '{} vs {}'.format(num, len(dataset))
	
	inds = torch.randperm(len(dataset))[:num].numpy() if shuffle else np.arange(num)
	return Subset_Dataset(dataset, inds)



# @AutoModifier('blurred')
class Blurred(DatasetBase):

	def __init__(self, config):

		prepend = config.pull('prepend-tfm', False)
		if prepend:
			raise NotImplementedError

		level = config.pull('blur-level', 5)
		assert level % 2 == 1, f'bad blur level: {level}'
		blur_type = config.pull('blur-type', 'uniform')

		assert blur_type == 'uniform', f'not implemented: {blur_type}'

		pad_type = config.pull('pad-type', 'reflect')
		padding = (level-1)//2

		super().__init__(config)

		self.prepend = prepend

		C = self.din[0]

		self.pad_type = pad_type
		self.padding = padding

		self.blur = nn.Conv2d(C, 1, groups=C, bias=False, kernel_size=level,
		                      padding=padding, padding_mode=pad_type)
		self.blur.weight.requires_grad = False
		self.blur.weight.copy_(1).div_(level**2)

		self.done = False
		if isinstance(self, Deviced):

			self.blur.to(self.device)

			key = config.pull('data_key', 'images')
			full = getattr(self, key)

			with torch.no_grad():
				full = self.blur(full).detach()

			setattr(self, key, full)

			self.done = True



	def __getitem__(self, item):

		if self.done:
			return item

		raise NotImplementedError

		x, *other = item





		pass




import os, sys

import argparse


MY_PATH = os.path.dirname(os.path.abspath(__file__))

if __name__ == '__main__':
        # path = os.path.join(train.DEFAULT_DATA_PATH, 'emnist')
        # torchvision.datasets.EMNIST(path, split='letters', download=True, train=True)

        # mp.freeze_support()
        # mp.set_start_method('spawn')
        # a = torch.arange(3)#.cuda()
        # p = mp.Process(target=t, args=(a,1,))
        # p.start()
        # p.join()


        print(sys.argv)

        parser = argparse.ArgumentParser(description='Process some integers.')
        parser.add_argument('test', type=str)
        parser.add_argument('--test2', type=str)

        args = parser.parse_known_args()
        print(args)


        # train.register_config('mnist', os.path.join(MY_PATH, 'mnist', 'config', 'base.toml'))
        #
        # print(train.config._config_registry)
        # C = train.get_config('mnist')
        # print(C.keys())




import sys, os
from pathlib import Path
from wrapt import ObjectProxy
import numpy as np
import torch
from torch.utils.data import Dataset as PytorchDataset
import h5py as hf

from omnibelt import unspecified_argument

from .. import util


class ExistingModes(util.Switchable):
	def __init_subclass__(cls, **kwargs):
		cls.available_modes = {'train'}
	
	@classmethod
	def add_existing_modes(cls, *modes):
		cls.available_modes.update(modes)
	
	@classmethod
	def get_available_modes(cls):
		return cls.available_modes

class DatasetBase(ExistingModes, util.Dimensions, util.Configurable, PytorchDataset):
	pass

class Sourced(DatasetBase):
	def __init__(self, A, dataroot=None, **kwargs):
		super().__init__(A, **kwargs)
		self.root = util.get_data_dir(A) if dataroot is None else dataroot
	
	def get_root(self):
		return self.root
	
class Batchable(DatasetBase): # you can select using a full batch
	def allow_batched(self):
		return True


class Deviced(util.Deviced, DatasetBase): # Full dataset is in memory, so it can be moved to GPU
	def __init__(self, *args, **kwargs):

		super().__init__(*args, **kwargs)
		self._buffers = []

	def register_buffer(self, name, buffer):
		'''
		By convention, all input buffers should be registered before output (label) buffers.
		This will register the buffer and set the `buffer` as an attribute of `self` with key `name`.

		:param name: name of this buffer
		:param buffer: torch.Tensor with data
		:return:
		'''
		self._buffers.append(name)
		self.__setattr__(name, buffer)

	def to(self, device):
		super().to(device)
		buffers = getattr(self, '_buffers', None)
		if buffers is not None:
			for name in self._buffers:
				try:
					val = getattr(self, name)
					if val is not None:
						new = val.to(device)
						if new is not None:
							self.__setattr__(name, new)
				except AttributeError:
					pass

class Downloadable(Sourced):
	@classmethod
	def download(cls, A, **kwargs):
		raise NotImplementedError

class MissingDatasetError(Exception):
	def __init__(self, name):
		super().__init__(f'Missing dataset {name} (it can be downloaded using the "download-dataset" script)')

class MissingFIDStatsError(Exception):
	def __init__(self, root, dim, modes, available=None):
		msg = f'no fid stats (dim={dim}, modes={modes}) found in: {root}'
		if available is not None:
			msg = msg + f' - available: {available}'
		super().__init__(msg)
		self.root = root
		self.modes = modes
		self.dim = dim
		self.available = available


class Image_Dataset(DatasetBase):

	def __init__(self, A, root=None, fid_ident=unspecified_argument, **other):
		if fid_ident is unspecified_argument:
			fid_ident = A.pull('fid_ident', None)
		super().__init__(A, **other)
		self.fid_ident = fid_ident

	def get_available_fid(self, name=None):
		if name is None:
			name = 'fid_stats.h5' if self.fid_ident is None else f'{self.fid_ident}_fid_stats.h5'
		if self.root is not None:
			path = self.root / name
			if path.is_file():
				with hf.File(path, 'r') as f:
					available = list(f.keys())
				
				available = [a.split('_')[:-1] for a in available if 'mu' in a]
				out = []
				for a in available:
					try:
						m, d = a
						out.append((int(d), m))
					except:
						pass
				return out

	def get_fid_stats(self, dim, *modes, name=None):
		if name is None:
			name = 'fid_stats.h5' if self.fid_ident is None else f'{self.fid_ident}_fid_stats.h5'
		available = None
		if self.root is not None:
			path = self.root / name
			if path.is_file():
				with hf.File(path, 'r') as f:
					available = f.keys()
					for mode in modes:
						key = f'{mode}_{dim}'
						if f'{key}_mu' in f:
							return f[f'{key}_mu'][()], f[f'{key}_sigma'][()]
	
		raise MissingFIDStatsError(self.root, dim, modes, available)


class List_Dataset(DatasetBase):

	def __init__(self, ls):
		self.data = ls

	def __getitem__(self, idx):
		return self.data[idx]

	def __len__(self):
		return len(self.data)
	
	
	
# region Wrappers

class DatasetWrapper(ObjectProxy):
	pass
	
class Indexed_Dataset(DatasetWrapper):
	def __getitem__(self, idx):
		return idx, self.__wrapped__[idx]


class Subset_Dataset(DatasetWrapper):

	def __init__(self, dataset, indices=None):
		super().__init__(dataset)
		self._self_indices = indices

		try:
			device = self.__wrapped__.get_device()
			if self._self_indices is not None:
				self._self_indices = self._self_indices.to(device)
		except AttributeError:
			pass

	def __getitem__(self, idx):
		return self.__wrapped__[idx] if self._self_indices is None else self.__wrapped__[self._self_indices[idx]]

	def __len__(self):
		return len(self.__wrapped__) if self._self_indices is None else len(self._self_indices)

class Repeat_Dataset(ObjectProxy):

	def __init__(self, dataset, factor):
		super().__init__(dataset)
		self._self_factor = factor
		self._self_num_real = len(dataset)
		self._self_total = self._self_factor * self._self_num_real
		print('Repeating dataset {} times'.format(factor))

	def __getitem__(self, idx):
		return self.__wrapped__[idx % self._self_num_real]

	def __len__(self):
		return self._self_total


class Format_Dataset(DatasetWrapper):

	def __init__(self, dataset, format_fn, format_args=None, include_original=False):
		super().__init__(dataset)

		self._self_format_fn = format_fn
		self._self_format_args = {} if format_args is None else format_args
		self._self_include_original = include_original

	def __getitem__(self, idx):

		sample = self.__wrapped__[idx]

		formatted = self._self_format_fn(sample, **self._self_format_args)

		if self._self_include_original:
			return formatted, sample

		return formatted

class Shuffle_Dataset(DatasetWrapper):

	def __init__(self, dataset):
		super().__init__(dataset)

		self._self_indices = torch.randperm(len(self.__wrapped__)).clone()

		try:
			device = self.__wrapped__.get_device()
			self._self_indices = self._self_indices.to(device).clone()
		except AttributeError:
			pass

	def __getitem__(self, idx):
		return self.__wrapped__[self._self_indices[idx]]

# endregion



import math
import omnifig as fig

from omnifig.config import ConfigList, ConfigIter


@fig.Component('pull')
def _config_pull(A):
	terms = A.pull('_terms', '<>_term', silent=True)
	args = A.pull('_args', {}, silent=True)
	return A.pull(*terms, **args)


def _process_single(term, A):
	op, val = None, term

	if isinstance(term, (tuple, list)):
		assert len(term) == 2, f'unknown term: {term}'
		op, val = term

	if isinstance(val, (tuple, list)):
		val = A.pull(*val, silent=True)
	elif isinstance(val, str):
		val = A.pull(val, silent=True)
	if op is None or op in {0, 1, 'id', 'i', 'x', ''}:
		return val

	if op in {'minv', '/', 'div'}:
		return 1 / val
	elif op in {'ainv', '-', 'sub'}:
		return -val
	raise Exception(f'unknown op: {op}')

@fig.Component('expr')
def _config_expression(A):  # TODO: boolean ops

	red = A.pull('_reduce', '<>_op', '+', silent=True)

	terms = A.pull('_terms', '<>_term', silent=True)

	if not isinstance(terms, (list, tuple)):
		terms = terms,

	vals = [_process_single(term, A) for term in terms]

	if red in {'+', 'add', 'sum'}:
		out = sum(vals)
	elif red in {'avg', 'average', 'mean'}:
		out = sum(vals) / len(vals)
	elif red in {'*', 'product', 'mul'}:
		out = math.prod(vals)
	elif red in {'%', 'mod'}:
		assert len(vals) == 2, f'bad red: {vals}'
		out = vals[0] % vals[1]
	elif red in {'//', 'idiv'}:
		assert len(vals) == 2, f'bad red: {vals}'
		out = vals[0] // vals[1]
	else:
		raise Exception(f'unknown reduction {red}')

	caste = A.pull('_caste', None, silent=True)

	if caste is None:
		return out
	elif caste == 'int':
		return int(out)
	elif caste == 'str':
		return str(out)
	elif caste == 'float':
		return float(out)
	raise Exception(f'unkonwn caste: {caste}')


@fig.Component('iter/repeat')
class Repeat_Iter(ConfigIter):
	def __init__(self, A):
		num = A.pull('_num', '<>_len')
		unpack = A.pull('_unpack', False)

		element = A['_element']

		auto = A.pull('auto_pull', True)
		include_key = A.pull('include_key', None)

		super().__init__(A, elements=ConfigList(data=num*[*element] if unpack else num*[element], parent=A),
		                 auto_pull=auto, include_key=include_key)


@fig.Component('copy')
def copy_config(A):

	src = A.pull('_src', None)

	mods = A.pull('_src_mod', None)

	if src is not None:
		src = A.pull(src, raw=True)

		A.push('_type', '_x_', silent=True)
		A.push('_src', '_x_', silent=True)
		A.push('_src_mod', '_x_', silent=True)

		updates = A.pull_self(raw=True)

		A.update(src)
		A.update(updates)

		if mods is not None:
			A.push('_mod', mods)

	return A.pull_self()





import sys, os, time
import numpy as np
import torch
import torch.multiprocessing as mp
from .. import util

stack_collate = util.make_collate(stack=True)
cat_collate = util.make_collate(stack=False)

class _Generator_Iterator(object):
	def __init__(self, gen, step_threshold, track_stats_tau,
	             step_cutoff, episode_cutoff,
	             step_limit, episode_limit, ):
		self.gen = gen
		self.total_steps = 0
		self.total_episodes = 0

		self.step_threshold = step_threshold
		self.step_cutoff = step_cutoff
		self.episode_cutoff = episode_cutoff
		self.stats = util.StatsMeter('rewards', 'len', tau=track_stats_tau) if track_stats_tau is not None else None

		self.step_limit = step_limit
		self.episode_limit = episode_limit
		self.steps_left = step_limit
		self.episodes_left = episode_limit
		#assert not (self.step_limit is None and self.episode_limit is None), 'must specify limit for iterator'

	def steps_generated(self):
		return self.total_steps
	def episodes_generated(self):
		return self.total_episodes

	def __len__(self):
		l = self.episode_limit if self.episode_limit is not None else self.step_limit
		return 0 if l is None else l # can be infinite

	def __iter__(self):
		return self

	def __next__(self):
		if (self.episodes_left is not None and self.episodes_left <= 0) \
				or (self.steps_left is not None and self.steps_left <= 0):
			raise StopIteration

		step_cutoff = self.steps_left if self.step_cutoff is None else self.step_cutoff
		if self.step_cutoff is not None and self.steps_left is not None:
			step_cutoff = min(step_cutoff, self.step_cutoff)

		episode_cutoff = self.episodes_left if self.episode_cutoff is None else self.episode_cutoff
		if self.episode_cutoff is not None and self.episodes_left is not None:
			episode_cutoff = min(episode_cutoff, self.episode_cutoff)

		E, S, rollouts = self.gen._rollout(stats=self.stats,
		                                   step_threshold=self.step_threshold,
		                                   episode_cutoff=episode_cutoff,
		                                   step_cutoff=step_cutoff)
		self.total_episodes += E
		self.total_steps += S
		if self.steps_left is not None:
			self.steps_left -= S
		if self.episodes_left is not None:
			self.episodes_left -= E

		return rollouts

class Generator(object):
	def __init__(self, env, agent, track_stats_tau=0.1,
	             drop_last_state=False, max_episode_length=None,
	             episode_cutoff=None, step_cutoff=None, # hard cutoff per iteration
	             step_threshold=None, # soft cutoff per iteration
	             step_limit=None, episode_limit=None): # overall budget (optional)
		super().__init__()

		self.env = env
		self.agent = agent

		self.stats = util.StatsMeter('rewards', 'len', tau=track_stats_tau) if track_stats_tau is not None else None
		self.drop_last = drop_last_state
		self.step_threshold = step_threshold
		self.episode_cutoff = episode_cutoff
		self.step_cutoff = step_cutoff
		self.max_episode_length = max_episode_length

		self.total_steps = 0
		self.total_episodes = 0

		self.step_limit = step_limit
		self.episode_limit = episode_limit

		assert not (self.step_threshold is None
		            and self.step_cutoff is None
		            and self.episode_cutoff is None), 'must specify iteration size in some way'

	def steps_generated(self):
		return self.total_steps
	def episodes_generated(self):
		return self.total_episodes

	def _rollout(self, stats=None, step_threshold=None, episode_cutoff=None, step_cutoff=None):
		if stats is None:
			stats = self.stats
		if step_threshold is None:
			step_threshold = self.step_threshold
		if episode_cutoff is None:
			episode_cutoff = self.episode_cutoff
		if step_cutoff is None:
			step_cutoff = self.step_cutoff

		states = []
		actions = []
		rewards = []
		info = []

		steps_generated = 0
		episodes_generated = 0

		with torch.no_grad():

			while (episode_cutoff is None or episodes_generated < episode_cutoff) \
					and (step_cutoff is None or steps_generated < step_cutoff) \
					and (step_threshold is None or steps_generated < step_threshold):

				S = []
				A = []
				AI = []
				EI = []
				R = []

				S.append(self.env.reset().view(1, -1))
				episodes_generated += 1

				done = False
				steps = 0
				while not done and (self.max_episode_length is None or steps < self.max_episode_length)\
						and (step_cutoff is None or steps_generated < step_cutoff):
					a, ai = self.agent.gen_action(S[-1])
					A.append(a)
					AI.append(ai)

					s, r, done, *einfo = self.env.step(a)
					ei = einfo[0] if len(einfo) else {}
					
					steps += 1
					steps_generated += 1

					s = s.view(1, -1)
					EI.append(ei)
					S.append(s)
					R.append(r)

				if self.drop_last:
					S = S[:-1]

				S = torch.cat(S)
				A = torch.cat(A)
				R = torch.stack(R)
				AI = cat_collate(AI)
				EI = cat_collate(EI)

				if self.stats is not None:
					stats.update('len', A.size(0))
					stats.update('rewards', R.sum())

				states.append(S)
				actions.append(A)
				info.append(AI)
				info[-1].update(EI)  # combine all info
				rewards.append(R)

			info = stack_collate(info)
			rollout = {
				'states': stack_collate(states),
				'actions': stack_collate(actions),
				'rewards': stack_collate(rewards),
			}
			rollout.update(info)

		self.total_episodes += episodes_generated
		self.total_steps += steps_generated

		return episodes_generated, steps_generated, rollout

	def __iter__(self):
		return _Generator_Iterator(self, step_threshold=self.step_threshold,
		                           track_stats_tau=self.stats.tau if self.stats is not None else None,
	             step_cutoff=self.step_cutoff, episode_cutoff=self.episode_cutoff,
	             step_limit=self.step_limit, episode_limit=self.episode_limit, )

	def __call__(self, *args, **kwargs):
		return self._rollout(*args, **kwargs)[-1]




from .. import util

# TODO

class Evaluator(util.Configurable):
	
	def compute(self, *args, **kwargs):
		out = self._compute(*args, **kwargs)
		if isinstance(out, dict):
			scores = {score:out.get(score, None) for score in self.get_scores()}
			results = {result:data for result, data in out.items() if result not in scores}
			out = scores, results
		return out
	
	def _compute(self, *args, **kwargs):
		raise NotImplementedError
	
	def get_scores(self):
		return []
	
	def get_results(self):
		return []
	
	pass



class EvaluationManager(util.StatsClient):
	
	
	
	pass






import torch
from torch import nn

import omnifig as fig


# from .. import framework as fm
from ..op import framework as fm
from .nets import MultiLayer

#
#
# class StorablePrior(fm.FunctionBase):
# 	def __init__(self, din=None, dout=None, prior_dim=None):
# 		super().__init__(din, dout)
# 		self.prior_dim = prior_dim
#
# 		self.volatile.prior = None
#
# 	def get_style_dim(self):
# 		return self.prior_dim
#
# 	def set_prior(self, p):
# 		self.volatile.prior = p
#
# 	def check_stored(self, N=1):
# 		if self.volatile.prior is not None:
# 			p = self.volatile.prior
# 			self.volatile.prior = None
# 			assert p.size(0) == N
# 			return p
# 		return self.sample_prior(N)
#
# 	def sample_prior(self, N=1):
# 		return torch.randn(N, self.prior_dim, device=self.device)
#
# @fig.Component('adain')
# class AdaIN(StorablePrior, fm.FunctionBase):
#
# 	def __init__(self, A):
# 		channels = None
# 		din = A.pull('din', None)
# 		dout = None
# 		if din is None:
# 			dout = A.pull('dout', None)
#
# 			if dout is None:
# 				channels = A.pull('channels')
# 			else:
# 				din = dout
# 				channels = dout[0]
# 		else:
# 			dout = din
#
# 			channels = dout[0]
#
# 		assert channels is not None, 'no info'
#
# 		norm = A.pull('normalize', False)
#
# 		style_dim = A.pull('style_dim', '<>latent_dim')
#
# 		A.push('net._type', 'mlp', overwrite=False, silent=True)
# 		A.push('net.din', style_dim, silent=True)
# 		A.push('net.dout', channels*2, silent=True)
# 		net = A.pull('net')
#
# 		super().__init__(din, dout, style_dim)
#
# 		self.feature_dim = channels
#
# 		self.net = net
#
# 		self.normalize = nn.InstanceNorm2d(channels, affine=False) if norm else None
#
# 	# def extra_repr(self):
# 	# 	return f'style={self.get_style_dim()}, normalize={self.normalize is not None}'
#
# 	def process_prior(self, p):
# 		return self.net(p)
#
# 	def process_features(self, x):
# 		return x if self.normalize is None else self.normalize(x)
#
# 	def forward(self, x, p=None):
#
# 		c = self.process_features(x)
#
# 		if p is None:
# 			p = self.check_stored(x.size(0))
# 		q = self.process_prior(p)
# 		m = q.narrow(1, 0, self.feature_dim).unsqueeze(-1).unsqueeze(-1)
# 		s = q.narrow(1,self.feature_dim, self.feature_dim).unsqueeze(-1).unsqueeze(-1).exp()
#
# 		return s*c + m
#
#
# @fig.Component('style')
# class StyleModel(MultiLayer, StorablePrior):
#
# 	def _create_layers(self, A):
#
# 		split_style = A.pull('split-style', False)
#
# 		layers = super()._create_layers(A)
#
# 		branches = [layer for layer in layers if isinstance(layer, StorablePrior)]
# 		bdims = [branch.get_style_dim() for branch in branches]
#
# 		if split_style:
# 			style_dim = sum(bdims)
# 		else:
# 			style_dim = bdims[0]
# 			assert all(b == style_dim for b in bdims)
#
# 		self.split_style = split_style
# 		self.branches = branches
# 		self.branch_dims = bdims
# 		self.style_dim = style_dim
# 		self.prior_dim = style_dim
#
# 		return layers
#
# 	def forward(self, x, z=None):
#
# 		if z is None:
# 			z = self.check_stored(x.size(0))
#
# 		if self.split_style:
# 			priors = z.split(self.branch_dims, dim=1)
# 			for branch, prior in zip(self.branches, priors):
# 				branch.set_prior(prior)
# 		else:
# 			for branch in self.branches:
# 				branch.set_prior(z)
#
# 		return super().forward(x)
#








from humpack.farming import Farmer, make_ghost, Replicator, Parallelizer, replicate, Cloner

# import sys, os
# import torch
# import numpy as np
# import torch.multiprocessing as mp
# import itertools
# import traceback

# class ExceptionWrapper(object):
# 	r"""Wraps an exception plus traceback to communicate across threads"""
# 	def __init__(self, exc_info):
# 		# It is important that we don't store exc_info, see
# 		# NOTE [ Python Traceback Reference Cycle Problem ]
# 		self.exc_type = exc_info[0]
# 		self.exc_msg = "".join(traceback.format_exception(*exc_info))
#
# def _worker_loop(fn, private_args, in_queue, out_queue, unique_args={}, init_fn=None):
# 	torch.set_num_threads(1)
#
# 	output_args = None
# 	if init_fn is not None:
# 		args = private_args.copy()
# 		args.update(unique_args)
# 		try:
# 			output_args = init_fn(**args)
# 		except Exception:
# 			print('failed')
# 			out_queue.put(ExceptionWrapper(sys.exc_info()))
# 			return
#
# 	while True:
# 		all_args = private_args.copy()
# 		all_args.update(unique_args)
# 		if output_args is not None:
# 			all_args.update(output_args)
#
# 		args = in_queue.get()
# 		if args is None:
# 			break
# 		try:
# 			shared_args, volatile_args = args
# 			all_args.update(shared_args)
# 			all_args.update(volatile_args)
# 			output = fn(**all_args)
# 		except Exception:
# 			out_queue.put(ExceptionWrapper(sys.exc_info()))
# 		else:
# 			out_queue.put(output)
#
# class Farmer(object):
# 	'''
# 	Farms computation (of functions with both private and shared args) to other processes
# 	4 types of arguments (all of which are dicts):
# 	- shared_args = arguments to be sent to workers when dispatched from master process
# 	- private_args = arguments that all workers own privately (eg. data you dont want to pass in each dispatch)
# 	- unique_worker_args = arguments that each worker owns privately (eg. random seeds)
# 	- volatile_args = arguments generated and sent for each dispatch
#
# 	'''
# 	def __init__(self, fn, shared_args={}, private_args={}, unique_worker_args=None, volatile_gen=None,
# 	             init_fn=None, num_workers=0, timeout=20, waiting=None, auto_dispatch=True):
# 		'''
#
# 		:param fn:
# 		:param shared_args:
# 		:param private_args:
# 		:param unique_worker_args:
# 		:param volatile_gen:
# 		:param init_fn:
# 		:param num_workers:
# 		:param timeout:
# 		:param waiting:
# 		:param auto_dispatch:
# 		'''
#
# 		if unique_worker_args is not None:
# 			try:
# 				num_workers = len(unique_worker_args)
# 			except TypeError:
# 				pass
#
# 		self.num_workers = num_workers
# 		self.shared_args = shared_args
# 		self.volatile_gen = volatile_gen
# 		self.timeout = timeout
# 		self.workers = None
#
# 		self.in_queue = mp.Queue()
# 		self.outstanding = 0
# 		self.auto_dispatch = auto_dispatch
#
# 		self.dispatch_num = 1
#
# 		if num_workers > 0:
# 			if unique_worker_args is not None:  # list of dicts
# 				assert len(unique_worker_args) == num_workers
# 			else:
# 				unique_worker_args = [{}] * num_workers
#
# 			self.out_queue = mp.Queue()
# 			self.workers = [
# 				mp.Process(target=_worker_loop, args=(fn, private_args, self.in_queue, self.out_queue, u, init_fn))
# 				for i, u in enumerate(unique_worker_args)]
#
# 			for w in self.workers:
# 				w.daemon = True  # ensure that the worker exits on process exit
# 				w.start()
#
# 			if waiting is None:
# 				waiting = num_workers if auto_dispatch else 0
# 			self._dispatch(waiting)
#
# 		else:
# 			self.fn = fn
# 			self.args = private_args.copy()
# 			if init_fn is not None:
# 				output_args = init_fn(**private_args)
# 				if output_args is not None:
# 					self.args.update(output_args)
#
# 	def _get_volatile_args(self):
# 		if self.volatile_gen is not None:
# 			return next(self.volatile_gen)
# 		return {}
#
# 	def _dispatch(self, n=1, args=None):
# 		if args is None:
# 			args = self.shared_args
# 		for _ in range(n):
# 			try:
# 				self.in_queue.put((args, self._get_volatile_args()))
# 				self.outstanding += 1
# 			except StopIteration:
# 				pass
#
# 	def dispatch(self, **kwargs):
# 		self._dispatch(self.dispatch_num, args=kwargs.update(self.shared_args))
#
# 	def __len__(self):
# 		return self.outstanding
#
# 	def __iter__(self):
# 		return self
#
# 	def __next__(self):
# 		if self.auto_dispatch:
# 			self._dispatch()
# 		if self.outstanding == 0:
# 			raise StopIteration
# 		if self.workers is None: # apply fn in this process
# 			args = self.args.copy()
# 			shared_args, volatile_args = self.in_queue.get()
# 			self.outstanding -= 1
# 			args.update(shared_args)
# 			args.update(volatile_args)
# 			return self.fn(**args)
# 		output = self.out_queue.get(timeout=self.timeout)
# 		self.outstanding -= 1
# 		if isinstance(output, ExceptionWrapper):
# 			try:
# 				raise output.exc_type(output.exc_msg)
# 			except:
# 				print('***ERROR: Exception of type {} occurred'.format(output.exc_type))
# 				raise Exception(output.exc_msg)
# 				#quit()
# 		return output
#
# 	def __del__(self):
# 		if self.workers is not None:
# 			for _ in self.workers:
# 				self.in_queue.put(None)
#
#
# def make_ghost(source, execute=None):
# 	'''
# 	upon anything done to a ghost instance it will check the 'source' for the correct behavior
# 	and will call 'execute' with the function and args
#
# 	:param execute: should be callable with signature: execute(fn, args=[], kwargs={})
# 	:param source: class which contains the functionality
# 	:return: Ghost object which can be used to execute functions from 'source' in 'parent'
# 	'''
#
# 	if execute is None:
# 		def execute(fn, args=[], kwargs={}):
# 			return fn(*args, **kwargs)
#
# 	def make_listener(fn):
# 		def listener(*args, **kwargs):
# 			return execute(fn, args, kwargs)
# 		return listener
#
# 	class Ghost(object):
# 		def __getattr__(ignore, item):
# 			#print('rpl', item)
# 			if hasattr(source, item):  # any method in obj_type
# 				return make_listener(getattr(source, item))
#
# 			return execute(getattr, args=[item])
#
# 		# NOTE: use r.__len__()
# 		#def __len__(ignore):
# 		#	raise Exception('Python error: len(r) doesn\'t work on replicas, use r.__len__() instead')
#
# 		def __setattr__(self, key, value):
# 			execute(getattr(source, '__setattr__'), args=[key, value])
#
# 		def __delattr__(self, item):
# 			execute(getattr(source, '__delattr__'), args=[item])
#
# 		def __getitem__(ignore, item):
# 			return execute(getattr(source, '__getitem__'), args=[item])
#
# 		def __iter__(self):
# 			return itertools.zip_longest(*execute(getattr(source, '__iter__')))
#
# 		def __add__(ignore, other):
# 			return execute(getattr(source, '__add__'), args=[other])
#
# 	return Ghost()
#
#
# # Init fn for any replica (used by Replicator, Parallelizer, and Cloner) - creates instance of replicated object
# def _replica_init(obj_type, init_args, init_kwargs, unique_init_kwargs={}):
# 	init_kwargs.update(unique_init_kwargs)
# 	return {'obj': obj_type(*init_args, **init_kwargs)}
#
# # Run fn for any replica (used by Replicator, Parallelizer, and Cloner) - applies function to be executed to instance
# def _replica_run(obj, fn, args, kwargs, **other_args):
# 	try:
# 		return fn(obj, *args, **kwargs)
# 	except Exception as e:
# 		return e
#
# class Replicator(object):  # see 'replicate' function below
# 	def __init__(self, obj_type, replicas=None, unique_init_kwargs=None, init_args=[], init_kwargs={},
# 	             timeout=20, collate=True):
#
# 		assert replicas is not None or unique_init_kwargs is not None, 'not sure how many replicas to make'
# 		if replicas is None:
# 			replicas = len(unique_init_kwargs)
# 		self.replicas = replicas
#
# 		replica_init_args = {
# 			'obj_type': obj_type,
# 			'init_args': init_args,
# 			'init_kwargs': init_kwargs,
# 		}
#
# 		if replicas > 0:
# 			if unique_init_kwargs is not None:  # list of dicts
# 				assert len(unique_init_kwargs) == replicas
# 			else:
# 				unique_init_kwargs = [{}] * replicas
#
# 			self.in_queues = np.array([mp.Queue() for _ in range(replicas)])
# 			self.out_queues = np.array([mp.Queue() for _ in range(replicas)])
# 			self.workers = [
# 				mp.Process(target=_worker_loop, args=(
# 				_replica_run, replica_init_args, in_q, out_q, {'unique_init_kwargs': unique}, _replica_init))
# 				for i, (unique, in_q, out_q) in enumerate(zip(unique_init_kwargs, self.in_queues, self.out_queues))]
#
# 			for w in self.workers:
# 				w.daemon = True  # ensure that the worker exits on process exit
# 				w.start()
#
# 		else:  # creates an invisible wrapper
# 			assert unique_init_kwargs is None
# 			self.obj = _replica_init(**replica_init_args)['obj']
#
# 		self.obj_type = obj_type
# 		self.collate = collate
# 		self._idx = None
# 		self.timeout = timeout
#
# 	def __len__(self):
# 		return self.replicas
#
# 	def _idx_execute(self, sel):
#
# 		if len(sel) == 0:
# 			return make_ghost(self.obj_type, self._execute)
#
# 		options = np.arange(self.replicas)
#
# 		idx = []
# 		for s in sel:
# 			new = options[s]
# 			try:
# 				idx.extend(new)
# 			except:
# 				idx.append(new)
#
# 		def execute_idx(fn=None, args=[], kwargs={}):
# 			return self._execute(fn, args, kwargs, idx=idx)
# 		return make_ghost(self.obj_type, execute_idx)
#
# 	def __call__(self, *sel):
# 		return self._idx_execute(sel)
#
# 	def __getitem__(self, sel):
# 		if isinstance(sel, (int, slice)):
# 			sel = [sel]
# 		return self._idx_execute(sel)
#
# 	def _execute(self, fn=None, args=[], kwargs={}, idx=None):
#
# 		shared_args = {
# 			'args': args,
# 			'kwargs': kwargs,
# 			'fn': fn,
# 		}
#
# 		# dispatch job
# 		if self.replicas == 0:
# 			#print(fn, args, kwargs)
# 			return fn(self.obj, *args, **kwargs)
#
# 		in_queues = self.in_queues
# 		out_queues = self.out_queues
#
# 		if idx is not None:
# 			in_queues = in_queues[idx]
# 			out_queues = out_queues[idx]
#
# 		for in_q in in_queues:
# 			in_q.put((shared_args, {}))
#
# 		output = []
# 		for out_q in out_queues:
# 			output.append(out_q.get(timeout=self.timeout))
# 			if isinstance(output[-1], ExceptionWrapper):
# 				raise output.exc_type(output.exc_msg)
#
# 		# collate output
# 		if self.collate and isinstance(output[0], tuple):
# 			output = ([o[i] for o in output] for i in range(len(output[0])))
#
# 		return output
#
# class Parallelizer(Replicator): # see 'replicate' function below
# 	#def __init__(self, *args, **kwargs):
# 	#	super(Parallelizer, self).__init__(*args, **kwargs)
#
# 	def _execute(self, fn=None, args=[], kwargs={}, idx=None):
#
# 		# dispatch job
# 		if self.replicas == 0:
# 			# print(fn, args, kwargs)
# 			return fn(self.obj, *args, **kwargs)
#
# 		in_queues = self.in_queues
# 		out_queues = self.out_queues
#
# 		if idx is not None:
# 			in_queues = in_queues[idx]
# 			out_queues = out_queues[idx]
#
# 		jlen = len(in_queues)
#
# 		if len(args) > 0:
# 			assert len(args[0]) == jlen
# 			args = zip(*args)
# 		else:
# 			args = [[]] * jlen
#
# 		if len(kwargs) > 0:
# 			# uncollate kwargs
# 			kwargs = [{k: v[i] for k, v in kwargs.items()} for i in range(jlen)]
# 		else:
# 			kwargs = [{}] * jlen
#
# 		for in_q, a, k in zip(in_queues, args, kwargs):
# 			in_q.put(({'args': a, 'kwargs': k, 'fn': fn}, {}))
#
# 		output = []
# 		for out_q in out_queues:
# 			output.append(out_q.get(timeout=self.timeout))
# 			if isinstance(output[-1], ExceptionWrapper):
# 				raise output.exc_type(output.exc_msg)
#
# 		# collate output
# 		if self.collate and isinstance(output[0], tuple):
# 			output = ([o[i] for o in output] for i in range(len(output[0])))
#
# 		return output
#
# def replicate(*args, separate_args=False, ghost=False, **kwargs):
# 	'''
# 	Creates replica objects for multiprocessing. Each process contains a unique instance of the same class ('obj_type').
#
# 	There are 2 types of managers: Replicators and Parallelizers (chosen using 'separate_args')
# 	Replicator - all replicas take the same arguments when called
# 	Parallelizer - each replica takes different arguments (passed in as a list)
#
# 	Subsets of replicas may be addressed by indexing or calling the manager. If 'ghost' is true a ghost object will be
# 	returned in a addition to the manager which will apply anything done to that object to all replicas
#
# 	import time
# 	class Example:
# 		def f(self, i):
# 			time.sleep(0.5)
# 			print(i)
# 			return i+1
#
# 	replicator, ghost = replicate(Example, replicas=8, ghost=True)
#
# 	replicator(4).f(5) # executes f on replica 4
# 	replicator[0,2:6].f(5) # executes f on replicas 0 and 2:6
# 	replicator[0,0].f(5) # executes f on replica 0 twice
# 	replicator().f(5) # executes f on all replicas
# 	ghost.f(5) # executes f on all replicas
#
# 	# parallelizers have the same behavior except for each arg and kwarg a list the same length as the number of replicas to be executed must be passed in.
#
# 	:param args: for manager (check Replicator)
# 	:param separate_args: if true, then for each argument in the source function, a list of arguments must be passed one for each replica
# 	:param kwargs: for manager (check Replicator
# 	:return: manager, replicas (essentially a voodoo doll)
# 	'''
#
# 	Manager = Parallelizer if separate_args else Replicator
#
# 	manager = Manager(*args, **kwargs)
#
# 	if not ghost:
# 		return manager
#
# 	replicas = make_ghost(manager.obj_type, manager._execute)
#
# 	return manager, replicas # anything done to the replicas object will be applied in parallel to all replicas, replicator holds info about replicas
#
#
# class Cloner(Farmer):
# 	'''
# 	executes any method in 'obj_type' N (default=num_workers) times with the same args using a pool of 'num_workers' workers
#
# 	unique init args can be passed to each worker, but args to all workers for each dispatch will be the same
#
# 	example:
#
# 	import time
# 	class Example:
# 		def f(self, i):
# 			time.sleep(0.5)
# 			print(i)
# 			return i+1
#
# 	clones = Cloner(Example, num_workers=4)
#
# 	out = clones(8).f(5)
# 	# takes about 1 sec and prints "5" 8 times
# 	# 'out' now contains [6, 6, 6, 6, 6, 6, 6, 6]
#
# 	'''
# 	def __init__(self, obj_type, default_N=None, init_args=[], init_kwargs={}, unique_worker_args=None,
# 	             num_workers=0, collate=True, timeout=20):
#
# 		assert num_workers is not None or unique_worker_args is not None, 'must specify how many workers to use'
#
# 		if unique_worker_args is not None:
# 			unique_worker_args = [{'unique_init_kwargs': unique} for unique in unique_worker_args]
# 			num_workers = len(unique_worker_args)
#
# 		worker_init_args = {
# 			'obj_type': obj_type,
# 			'init_args': init_args,
# 			'init_kwargs': init_kwargs,
# 		}
#
# 		super(Cloner, self).__init__(fn=_replica_run, init_fn=_replica_init,
# 		                             private_args=worker_init_args, unique_worker_args=unique_worker_args,
# 		                             num_workers=num_workers, auto_dispatch=False, timeout=timeout)
# 		self.default_N = max(num_workers, 1) if default_N is None else default_N
# 		self.obj_type = obj_type
# 		self.collate = collate
#
# 		#self.clones = make_ghost(self.obj_type, self._execute)
#
# 	def __call__(self, N=None):
# 		def execute_N(fn=None, args=[], kwargs={}):
# 			return self._execute(fn, args, kwargs, N)
# 		return make_ghost(self.obj_type, execute_N)
#
# 	def _execute(self, fn=None, args=[], kwargs={}, N=None): # execution should not change the state of the clone
#
# 		self.shared_args = {
# 			'args': args,
# 			'kwargs': kwargs,
# 			'fn': fn,
# 		}
#
# 		if N is None:
# 			N = self.default_N
#
# 		# dispatch job
# 		self._dispatch(N)
#
# 		# collect responses
# 		output = [out for out in self]
#
# 		# collate output
# 		if self.collate and isinstance(output[0], tuple):
# 			output = ([o[i] for o in output] for i in range(len(output[0])))
#
# 		return output




import os
import numpy as np
import cv2
from io import BytesIO
from PIL import Image
import torch
from torch.utils.data import DataLoader, Dataset
# from torch.utils.data.dataloader import re, numpy_type_map, _use_shared_memory, string_classes, int_classes#,
import re
try:
	from torch.utils.data.dataloader import container_abcs
except:
	pass

from .misc import TreeSpace
import random
import collections
import itertools
import h5py as hf

import omnifig as fig

import torch
from torch import nn
from torch.nn import functional as F


class make_infinite(DataLoader):
	def __init__(self, loader, extractor=None):
		assert len(loader) >= 1, 'has no len'
		self.loader = loader
		self.itr = None
		self.extractor = extractor
		
		self.cached = None
	
	def __len__(self):
		return len(self.loader)
	
	def end(self):
		self.itr = None
		self.empty()
	
	def empty(self):
		self.cached = None
	
	def demand(self, N, extract=None, merge=None):
		'''
		
		:param N:
		:param extract: callable input: raw batch; output: mergable data with len()
		:param merge: callable input: list of extracted batches; output: whatever is expected
		:return:
		'''
		
		if extract is None:
			extract = self.extractor
		
		missing = N
		batches = []
		
		while missing > 0:
			if self.cached is None:
				x = next(self)
				# by default, assume batch is a tuple and pull first element
				x = x[0] if extract is None else extract(x)
			else:
				x = self.cached
				self.cached = None
			
			if len(x) > missing:
				self.cached = x[missing:]
				x = x[:missing]
			missing -= len(x)
			batches.append(x)
		
		return (torch.cat(batches),) if merge is None else merge(batches)
	
	def __next__(self):
		if self.itr is not None:
			try:
				return next(self.itr)
			except StopIteration:
				pass
		return next(iter(self))
	
	def __iter__(self):
		self.itr = iter(self.loader)
		return self
	


def get_h5_data(path, key, idx=None):
	with hf.File(path, 'r') as f:
		return f[key].value if idx is None else f[key][idx]


def make_np_wrapped_env(env_type):
	def create(*args, **kwargs):
		return Numpy_Env_Wrapper(env_type(*args, **kwargs))
	return create
class Numpy_Env_Wrapper(object):
	def __init__(self, env, device='cpu'):
		self._env = env
		self.device = device
		
	def to(self, d):
		self.device = d
	def cpu(self):
		self.device = 'cpu'
	def cuda(self, i=0):
		self.device = 'cuda:{}'.format(i)

	def __getattr__(self, item):
		if item not in self.__dict__:
			return self._env.__getattribute__(item)
		return super(Numpy_Env_Wrapper, self).__getattribute__(item)

	def reset(self, state=None):
		if state is None:
			return torch.from_numpy(self._env.reset()).float().to(self.device)
		return torch.from_numpy(self._env.reset(state.detach().cpu().numpy())).float().to(self.device)

	def step(self, action):
		ns, r, d, info = self._env.step(action.detach().cpu().numpy())
		ns = torch.from_numpy(ns).float().to(self.device)
		r = torch.tensor(r).float().to(self.device)
		return ns, r, d, info

	def render(self, *args, **kwargs):
		return self._env.render(*args, **kwargs) # no wrapping here



def batches(iterable, n): # batches an iterable (batches are not lazy)
	iterable = iter(iterable)
	while True:
		tup = tuple(itertools.islice(iterable, 0, n))
		if tup:
			yield tup
		else:
			break


def fill_in(src, new, locs, add=True, inplace=False):
	'''fill in `new` in `src` at pixel coords `locs` (B,2) (batched)'''

	squeeze = False
	if len(src.shape) == 3:
		B, H, W = src.shape
		C = 1
		squeeze = True
		src.unsqueeze(1)
		if len(new.shape) == 3:
			new = new.unsqueeze(1)
	else:
		assert len(src.shape) == 4

		B, C, H, W = src.shape

		if C != 1:
			raise NotImplementedError

	U, V = new.shape[-2:]

	u, v = locs.t()

	vals = new.view(-1)
	src = src.view(-1)

	shift = torch.arange(0, U * V).fmod(V).eq(0)
	shift[0] = 0
	shift = shift.int().mul(W - V).cumsum(0)

	pos = torch.arange(U * V * C).unsqueeze(0)

	idx = u * W + v

	inds = (shift + pos + (H * W * C * torch.arange(B) + idx).unsqueeze(1)).view(-1)

	if inplace:

		if add:
			src.index_add_(0, inds, vals)
		else:
			src.index_copy_(0, inds, vals)
		out = src

	else:

		if add:
			out = src.index_add(0, inds, vals)
		else:
			out = src.index_copy(0, inds, vals)

		out = out.view(B, C, H, W)

	if squeeze:
		out.squeeze(1)
	return out


def to_one_hot(idx, max_idx=None):
	if max_idx is None:
		max_idx = idx.max()
	dims = (max_idx,)
	if idx.ndimension() >= 1:
		if idx.size(-1) != 1:
			idx = idx.unsqueeze(-1)
		dims = idx.size()[:-1] + dims
	return torch.zeros(*dims).to(idx.device).scatter_(-1, idx.long(), 1)

# linear - discretize continuous values into N classes
# TODO: implement log space
def discretize(input, N, range=None):

	if range is None:
		range = input.min_val(), input.max()

	input = input.clamp(*range)
	input -= range[0]

	return input.mul((N-1)/(range[1]-range[0])).round().long()

## load image

def load_images(*paths, root=None, channel_first=False, conv_brg=True, **flags):
	assert not len(flags)
	for path in paths:
		if root is not None:
			path = os.path.join(root, path)
		img = cv2.imread(path, cv2.IMREAD_UNCHANGED)
		if conv_brg and len(img.shape) == 3 and img.shape[-1] in {3,4}:
			img = img[..., ::-1].copy()
		if channel_first:
			img = img.transpose(2,0,1)
		yield img

def read_raw_bytes(path, root=None):
	if root is not None:
		path = os.path.join(root, path)
	with open(path, 'rb') as f:
		return f.read()


#########################
# Util to convert images to strings for compressed saving

def rgb_to_str(img, compression=5):
	assert img.shape[2] == 3
	assert img.dtype == np.uint8

	return cv2.imencode('.png', img, [cv2.IMWRITE_PNG_COMPRESSION, compression])[1].tostring()


def float32_img_to_str(img, compression=5):
	H, W = img.shape
	assert img.dtype == np.float32

	img.dtype = np.uint8
	img.shape = (H, W, 4)

	s = cv2.imencode('.png', img, [cv2.IMWRITE_PNG_COMPRESSION, compression])[1].tostring()

	img.shape = (H, 4 * W)
	img.dtype = np.float32

	return s


def byte_img_to_str(img, compression=5):
	assert img.dtype == np.uint8

	return cv2.imencode('.png', img, [cv2.IMWRITE_PNG_COMPRESSION, compression])[1].tostring()


def str_to_rgb(s):
	return cv2.imdecode(np.frombuffer(s, np.uint8), -1)


def str_to_float32_img(s):
	img = cv2.imdecode(np.frombuffer(s, np.uint8), -1)

	H, W = img.shape[:2]

	img = img.reshape(H, W * 4)
	img.dtype = np.float32

	return img


def str_to_byte_img(s):
	return cv2.imdecode(np.frombuffer(s, np.uint8), -1)


def jpeg_to_str(path):
	with open(path, 'rb') as f:
		return f.read()

def str_to_jpeg(s, ret_PIL=False):
	img = Image.open(BytesIO(s))
	if ret_PIL:
		return img
	return np.array(img)


# Wrapper to create a DataLoader with NS_collate
# def loader(dataset, def_type=None, multi_agent=False, **kwargs): # switches default_collate to NS_collate
# 	if 'collate_fn' not in kwargs:
# 		kwargs['collate_fn'] = make_collate(def_type, multi_agent)
# 	if 'batch_size' in kwargs and kwargs['batch_size'] is None:
# 		kwargs['batch_size'] = len(dataset)
# 	return DataLoader(dataset, **kwargs)


### old

# multi traj gen: [N] x {NS} x [A]  --> [A] x {NS} x [N]
#
#
# def make_collate(stack=True):
# 	merge_fn = torch.stack if stack else torch.cat
#
# 	def collate(batch):
# 		r"""Puts each data field into a tensor with outer dimension batch size"""
#
# 		error_msg = "batch must contain tensors, numbers, dicts or lists; found {}"
# 		elem_type = type(batch[0])
# 		if isinstance(batch[0], torch.Tensor):
# 			out = None
# 			if True:
# 				# If we're in a background process, concatenate directly into a
# 				# shared memory tensor to avoid an extra copy
# 				numel = sum([x.numel() for x in batch])
# 				storage = batch[0].storage()._new_shared(numel)
# 				out = batch[0].new(storage)
# 			try:
# 				return merge_fn(batch, 0, out=out)
# 			except RuntimeError:
# 				return batch
# 		elif elem_type.__module__ == 'numpy' and elem_type.__name__ != 'str_' \
# 				and elem_type.__name__ != 'string_':
# 			elem = batch[0]
# 			if elem_type.__name__ == 'ndarray':
# 				# array of string classes and object
# 				if re.search('[SaUO]', elem.dtype.str) is not None:
# 					raise TypeError(error_msg.format(elem.dtype))
#
# 				return torch.stack([torch.from_numpy(b) for b in batch], 0)
# 			if elem.shape == ():  # scalars
# 				# py_type = float if elem.dtype.name.startswith('float') else int
# 				# return numpy_type_map[elem.dtype.name](list(map(py_type, batch)))
# 				return torch.tensor(batch)
# 		elif isinstance(batch[0], int):
# 			return torch.LongTensor(batch)
# 		elif isinstance(batch[0], float):
# 			return torch.DoubleTensor(batch)
# 		elif isinstance(batch[0], str):
# 			return batch
# 		elif isinstance(batch[0], container_abcs.Mapping):
# 			return {key: collate([d[key] for d in batch]) for key in batch[0]}
# 		elif isinstance(batch[0], container_abcs.Sequence):
# 			transposed = zip(*batch)
# 			return [collate(samples) for samples in transposed]
#
# 		raise TypeError((error_msg.format(type(batch[0]))))
#
# 	return collate
#
#
#
# def NS_make_collate(def_type=None, multi_agent=False):
#
# 	collate_fn = None
#
# 	if def_type is None:
# 		collate_fn = NS_list_collate
# 		return collate_fn
#
# 	def NS_list_type_collate(batch):
# 		r"""Puts each data field into a tensor with outer dimension batch size"""
#
# 		full = TreeSpace()
#
# 		for key in batch[0].keys():
# 			full[key] = []
# 			for sample in batch:
# 				data = NS_collate(sample[key]) if key == 'env_infos' else NS_collate([sample[key]])[0] # TODO: clean up
# 				if isinstance(data, torch.Tensor):
# 					data = data.type(def_type)
# 				full[key].append(data)
#
# 		return full  # returns NS containing collated paths
#
# 	collate_fn = NS_list_type_collate
#
# 	if multi_agent:
# 		if def_type is None:
# 			return NS_multi_list_collate
#
# 		def NS_multi_list_type_collate(batch):  # no default tensor type
# 			r"""Puts each data field into a tensor with outer dimension batch size"""
#
# 			num_agents = len(next(iter(batch[0].values())))
#
# 			full = [TreeSpace() for _ in range(num_agents)]  # a separate batch of paths for each agent
#
# 			for i, f in enumerate(full):
# 				for key in batch[0].keys():
# 					for sample in batch:
# 						if isinstance(sample[key], list):
# 							if isinstance(sample[key][0], dict):
# 								data = NS_collate(sample[key])
# 							else:
# 								data = NS_collate([sample[key][i]])[0]
# 								if isinstance(data, torch.Tensor):
# 									data = data.type(def_type)
# 						else:
# 							data = NS_collate([sample[key]])[0]
# 							if isinstance(data, torch.Tensor):
# 								data = data.type(def_type)
# 						if key not in f:
# 							f[key] = []
# 						f[key].append(data)
#
# 			#print(full[0].rewards)
#
# 			return full  # returns NS containing collated paths
#
# 		return NS_multi_list_type_collate
#
# 	return collate_fn
#
#
# def NS_multi_list_collate(batch):  # no default tensor type
# 	r"""Puts each data field into a tensor with outer dimension batch size"""
#
# 	num_agents = len(next(batch[0].values()))
#
# 	full = [TreeSpace() for _ in range(num_agents)] # a separate batch of paths for each agent
#
# 	for i, f in enumerate(full):
# 		for key in batch[0].keys():
# 			f[key] = [NS_collate([sample[key][i]])[0] for sample in batch]
#
# 	return full  # returns NS containing collated paths
#
# def NS_list_collate(batch): # no default tensor type
# 	r"""Puts each data field into a tensor with outer dimension batch size"""
#
# 	full = TreeSpace()
#
# 	if isinstance(batch, dict):
# 		full = TreeSpace(**{k:NS_list_collate(batch[k]) for k in batch.keys()})
# 		return full
#
# 	for key in batch[0].keys():
# 		full[key] = [NS_collate([sample[key]])[0] for sample in batch]
#
# 	return full # returns NS containing collated paths
#
# def NS_collate(batch):
# 	r"""Puts each data field into a tensor with outer dimension batch size"""
#
# 	error_msg = "batch must contain tensors, numbers, dicts or lists; found {}"
#
# 	if isinstance(batch, dict):
# 		return {k:NS_collate(batch[k]) for k in batch.keys()}
#
# 	elem_type = type(batch[0])
# 	if isinstance(batch[0], torch.Tensor):
# 		out = None
# 		if _use_shared_memory:
# 			# If we're in a background process, concatenate directly into a
# 			# shared memory tensor to avoid an extra copy
# 			numel = sum([x.numel() for x in batch])
# 			storage = batch[0].storage()._new_shared(numel)
# 			out = batch[0].new(storage)
# 		return torch.stack(batch, 0, out=out)
# 	elif elem_type.__module__ == 'numpy' and elem_type.__name__ != 'str_' \
# 			and elem_type.__name__ != 'string_':
# 		elem = batch[0]
# 		if elem_type.__name__ == 'ndarray':
# 			# array of string classes and object
# 			if re.search('[SaUO]', elem.dtype.str) is not None:
# 				raise TypeError(error_msg.format(elem.dtype))
#
# 			return torch.stack([torch.from_numpy(b) for b in batch], 0)
# 		if elem.shape == ():  # scalars
# 			py_type = float if elem.dtype.name.startswith('float') else int
# 			return numpy_type_map[elem.dtype.name](list(map(py_type, batch)))
# 	elif isinstance(batch[0], int_classes):
# 		return torch.LongTensor(batch)
# 	elif isinstance(batch[0], float):
# 		return torch.DoubleTensor(batch)
# 	elif isinstance(batch[0], string_classes):
# 		return batch
# 	elif isinstance(batch[0], TreeSpace):
# 		return TreeSpace(**{key: NS_collate([d[key] for d in batch]) for key in batch[0]})
# 	elif isinstance(batch[0], collections.Mapping):
# 		return {key: NS_collate([d[key] for d in batch]) for key in batch[0]}
# 	elif isinstance(batch[0], torch.distributions.Distribution):
# 		return [distrib for distrib in batch]
# 	elif isinstance(batch[0], collections.Sequence):
# 		transposed = zip(*batch)
# 		return [NS_collate(samples) for samples in transposed]
#
# 	raise TypeError((error_msg.format(type(batch[0]))))